import matplotlib.pyplot as plt 
import xarray as xr
import xarray.ufuncs as uf
import numpy as np
import warnings
import gsw
import traceback
from .COAsT import COAsT
from .NEMO import NEMO
from scipy import interpolate
from scipy.integrate import cumtrapz
from sklearn.neighbors import BallTree
from skimage import measure

# =============================================================================
# The contour module is a place for code related to contours only
# =============================================================================

class Contour:
    GRAVITY = 9.8 # m s^-2
    EARTH_ROT_RATE = 7.2921 * 10**(-5) # rad/s
    
    
    @staticmethod
    def get_contours(nemo: COAsT, contour_depth: int):
        '''
        A method to obtain the continuous isbobath contours within a supplied 
        Nemo domain as a set of y indices and x indices for the model grid.

        Parameters
        ----------
        nemo : COAsT
            The Nemo object containing the dataset with the 'bathymetry' variable
        contour_depth : int
            Depth of desired contours

        Returns
        -------
        List of 2d ndarrays
            Each item of the list contains a different continuous isobath contour as a
            2d ndarray of indicies, i.e. for each list item:
            contour[:,0] contains the y indices for the contour on the model grid
            contour[:,1] contains the x indices for the contour on the model grid

        '''
        contours = measure.find_contours( nemo.dataset.bathymetry.data, contour_depth )
        # The find_contours method returns indices that have been interpolated
        # between grid points so we must round and cast to integer 
        contours = [np.round(contour).astype(int) for contour in contours]
        return contours, len(contours)
    
    
    @staticmethod
    def plot_contour(nemo: COAsT, contour: np.ndarray):
        '''
        Quick plot method to plot a contour over a pcolormesh of the 
        model bathymetry

        Parameters
        ----------
        nemo : COAsT
            The Nemo object containing the dataset with the 'bathymetry' variable
        contour : 2d array
            contour[:,0] contains the y indices for the contour on the model grid
            contour[:,1] contains the x indices for the contour on the model grid
            i.e. contour = np.vstack((y_indices,x_indices)).T

        Returns
        -------
        None

        '''
        fig,ax=plt.subplots()
        lat = nemo.dataset.latitude[xr.DataArray(contour[:,0]), xr.DataArray(contour[:,1])]
        lon = nemo.dataset.longitude[xr.DataArray(contour[:,0]), xr.DataArray(contour[:,1])]
        
        nemo.dataset.bathymetry.where(nemo.dataset.bathymetry > 0, np.nan) \
            .plot.pcolormesh(y='latitude',x='longitude',ax=ax)
        ax.scatter(lon,lat, s=0.5, color='r')
        
        
    @staticmethod    
    def get_contour_segment(nemo: COAsT, contour: np.ndarray, start_coords: np.ndarray, end_coords: np.ndarray):
        '''
        Method that will take a contour from the list of contours generated by
        coast.Contour.get_contours() and trim it to start at supplied (lat,lon)
        coordinates and end at supplied (lat, lon) coordinates.

        Parameters
        ----------
        nemo : COAsT
            The Nemo object containing the dataset with the 'bathymetry' variable
        contour : numpy.ndarray
            contour[:,0] contains the y indices for the contour on the model grid
            contour[:,1] contains the x indices for the contour on the model grid
        start_coords : numpy.ndarray
            1d array containing [latitude,longitude] of the start point of the contour
        end_coords : numpy.ndarray
            1d array containing [latitude,longitude] of the end point of the contour

        Returns
        -------
        y_ind : numpy.ndarray
            y indices of the contour on the model grid
        x_ind : numpy.ndarray
            x indices of the contour on the model grid
        contour : numpy.ndarray
            For the convenience of plotting using coast.Contour.plot_contour()
            contour[:,0] = y_ind
            contour[:,1] = x_ind

        '''

        y_ind = contour[:,0]
        x_ind = contour[:,1]
        
        # Create tree of lat and lon on the pre-processed contour
        bt = BallTree( np.deg2rad( list( zip( nemo.dataset.latitude.values[y_ind, x_ind], 
                    nemo.dataset.longitude.values[y_ind, x_ind] ) ) ), metric='haversine' )

        # Get start and end indices for contour and subset accordingly
        start_idx = bt.query( np.deg2rad( [start_coords]) )[1][0][0]
        end_idx = bt.query( np.deg2rad([end_coords]) )[1][0][0]   
        if start_idx > end_idx:
            y_ind = y_ind[end_idx:start_idx+1] 
            x_ind = x_ind[end_idx:start_idx+1]
        else:
            y_ind = y_ind[start_idx:end_idx+1] 
            x_ind = x_ind[start_idx:end_idx+1]
            
        # Ensure that the start point is closer to southern boundary of domain.
        # If start and end point have same latitude then ensure start point is 
        # closer to the western boundary of the domain.
        if y_ind[0] > y_ind[-1]:
            y_ind = y_ind[::-1]
            x_ind = x_ind[::-1]
        elif y_ind[0] == y_ind[-1]:
            if x_ind[0] > x_ind[-1]:
                y_ind = y_ind[::-1]
                x_ind = x_ind[::-1]
                
        return y_ind, x_ind, np.vstack((y_ind,x_ind)).T
                        
    
    def __init__(self, nemo: COAsT, y_ind, x_ind, depth: int):        
        '''
        Class defining a Contour type, which is a 3d dataset of points between a point A and 
        a point B defining an isobath contour. The dataset has a time, depth and contour dimension. 
        The cotnour dimension defines the points along the contour.
        The supplied model Data is subsetted in its entirety along these dimensions and
        calculations can be performed on this dataset.
        
        Parameters
        ----------
        nemo : COAsT
            nemo object containing the model dataset.
        y_ind : numpy.ndarray
            1d array of y indices defining the contour on the model grid
        x_ind : numpy.ndarray
            1d array of x indices defining the contour on the model grid
        depth : int
            Depth of contour isobath
        '''
        try:
            if y_ind[0] > y_ind[-1]:
                raise ValueError("Start point of the contour " \
                                 "must be closer than the end point of the " \
                                 "contour to the southern boundary of the model " \
                                 "domain.")
            elif y_ind[0] == y_ind[-1]:
                if x_ind[0] > x_ind[-1]:
                    raise ValueError("Start and end points of the contour " \
                                     "have the same latitudes, the start point must " \
                                     "be the closer of the two points to the western " \
                                     "boundary of the model domain.")

        
            self.depth = depth
            self.y_ind, self.x_ind = self.process_contour( nemo.dataset, y_ind, x_ind )
            self.len = len(self.y_ind)
            self.filename_domain = nemo.filename_domain
            da_y_ind = xr.DataArray( self.y_ind, dims=['r_dim'] )
            da_x_ind = xr.DataArray( self.x_ind, dims=['r_dim'] )
            self.data_contour = nemo.dataset.isel(y_dim = da_y_ind, x_dim = da_x_ind)
        except ValueError:
            print(traceback.format_exc())
        
        
    def process_contour(self, ds: xr.Dataset, y_ind, x_ind):
        
        """ Redefine contour so that each point on the contour defined by
        y_ind and x_ind is seperated from its neighbours by a single index change 
        in y or x, but not both. 
        example: convert y_ind = [10,11], x_ind = [1,2] to y_ind = [10,10], x_ind = [1,2]
        or  y_ind = [10,11], x_ind = [1,1]
         
        Parameters
        ----------
        nemo : xarray.Dataset
            xarray Dataset from supplied nemo object
        y_ind : numpy.ndarray
            1d array of y indices defining the contour on the model grid
        x_ind : numpy.ndarray
            1d array of x indices defining the contour on the model grid

        Returns
        -------
        y_ind : numpy.ndarray
            processed y indices of the contour on the model grid
        x_ind : numpy.ndarray
            processed x indices of the contour on the model grid
         
        """
         
        try:
            y_ind = np.asarray(y_ind)
            x_ind = np.asarray(x_ind)
            # When replacing diagonal segments in the contour, pick the path that is
            # closest to the contour isobath depth
            option1 = uf.fabs(ds.bathymetry[xr.DataArray(y_ind+1), xr.DataArray(x_ind)] - self.depth)
            option0 = uf.fabs(ds.bathymetry[xr.DataArray(y_ind), xr.DataArray(x_ind+1)] - self.depth)
            add_new_y_point = xr.where(option1 <= option0, 1, 0 )
        
            spacing = np.abs( np.diff(y_ind) ) + np.abs( np.diff(x_ind) )
            if spacing.max() > 2:
                raise ValueError("The contour is not continuous. The contour must be defined on " 
                                 "adjacent grid points.")
            spacing[spacing!=2]=0
            doublespacing = np.nonzero( spacing )[0]
            for ispacing in doublespacing[::-1]:
                if add_new_y_point[ispacing]:
                    y_ind = np.insert( y_ind, ispacing+1, y_ind[ispacing+1] )
                    x_ind = np.insert( x_ind, ispacing+1, x_ind[ispacing] )
                else:
                    y_ind = np.insert( y_ind, ispacing+1, y_ind[ispacing] )
                    x_ind = np.insert( x_ind, ispacing+1, x_ind[ispacing+1] ) 
            
            # Remove any repeated points caused by the rounding of the indices
            nonrepeated_idx = np.nonzero( np.abs( np.diff(y_ind) ) + np.abs( np.diff(x_ind) ) )
    
            y_ind = np.concatenate( (y_ind[nonrepeated_idx], [y_ind[-1]]) )
            x_ind = np.concatenate( (x_ind[nonrepeated_idx], [x_ind[-1]]) )
    
            return (y_ind, x_ind)
        except ValueError:
            print(traceback.format_exc())
   
    
    def gen_z_levels(self, max_depth):
        ''' Generates a pre-defined 1d vertical depth coordinates,
        i.e. horizontal z-level vertical coordinates up to a supplied 
        maximum depth, 'max_depth' '''
        
        max_depth = max_depth + 650
        z_levels_0_50 = np.arange(0,55,5)
        z_levels_60_290 = np.arange(60,300,10)
        z_levels_300_600 = np.arange(300,650,50)
        z_levels_650_ = np.arange(650,max_depth+150,150)
        z_levels = np.concatenate( (z_levels_0_50, z_levels_60_290, 
                                    z_levels_300_600, z_levels_650_) )
        z_levels = z_levels[z_levels <= max_depth] 
        return z_levels
    

class Contour_f(Contour):
    '''
    Class defining a Contour type on the f-grid, which is a 3d dataset of points between a point A and 
    a point B defining an isobath contour. The dataset has a time, depth and contour dimension. 
    The contour dimension defines the points along the contour.
    The supplied model f-grid Data is subsetted in its entirety along these dimensions 
    within Contour_f.data_contour of type xarray.Dataset

    
    Parameters
    ----------
    nemo : COAsT
        f-grid nemo object containing the model dataset.
    y_ind : numpy.ndarray
        1d array of y indices defining the contour on the model grid
    x_ind : numpy.ndarray
        1d array of x indices defining the contour on the model grid
    depth : int
        Depth of contour isobath
    '''
    
    def __init__(self, nemo_f: COAsT, y_ind, x_ind, depth):
        super().__init__(nemo_f, y_ind, x_ind, depth)
        self.data_cross_flow = xr.Dataset()
     
        
    def calc_cross_contour_flow(self, nemo_u: COAsT, nemo_v: COAsT):
        '''
        Method that will calculate the flow across the contour and store this data
        within Contour_f.data_cross_flow, which is an xarray.Dataset. Specifically
        Contour_f.normal_velocities are the velocities across the contour
        (time, depth, position along contour) in m/s
        Contour_f.depth_integrated_normal_transport are the depth integrated 
        volume transports across the contour (time, position along contour) in Sv

        Parameters
        ----------
        nemo_u : COAsT
            The nemo object containing the model data on the u-grid.
        nemo_v : COAsT
            The nemo object containing the model data on the v-grid.

        Returns
        -------
        None.

        '''
                
        # subset the u and v datasets 
        da_y_ind = xr.DataArray( self.y_ind, dims=['r_dim'] )
        da_x_ind = xr.DataArray( self.x_ind, dims=['r_dim'] )
        u_ds = nemo_u.dataset.isel(y_dim = da_y_ind, x_dim = da_x_ind)
        v_ds = nemo_v.dataset.isel(y_dim = da_y_ind, x_dim = da_x_ind)
        
        dr_n = np.where(np.diff(self.y_ind)>0, np.arange(0,u_ds.r_dim.size-1), np.nan )
        dr_n = dr_n[~np.isnan(dr_n)].astype(int)
        dr_s = np.where(np.diff(self.y_ind)<0, np.arange(0,u_ds.r_dim.size-1), np.nan )
        dr_s = dr_s[~np.isnan(dr_s)].astype(int)
        dr_e = np.where(np.diff(self.x_ind)>0, np.arange(0,v_ds.r_dim.size-1), np.nan )
        dr_e = dr_e[~np.isnan(dr_e)].astype(int)
        dr_w = np.where(np.diff(self.x_ind)<0, np.arange(0,v_ds.r_dim.size-1), np.nan )
        dr_w = dr_w[~np.isnan(dr_w)].astype(int)
        
        # Note that subsetting the dataset first instead of subsetting each array seperately,
        # as we do here, is neater but significantly slower.
        self.data_cross_flow['normal_velocities'] = xr.full_like(u_ds.vozocrtx, np.nan)        
        self.data_cross_flow['normal_velocities'][:,:,dr_n] = u_ds.vozocrtx.data[:,:,dr_n+1]
        self.data_cross_flow['normal_velocities'][:,:,dr_s] = -u_ds.vozocrtx.data[:,:,dr_s]
        self.data_cross_flow['normal_velocities'][:,:,dr_e] = -v_ds.vomecrty.data[:,:,dr_e+1]
        self.data_cross_flow['normal_velocities'][:,:,dr_w] = v_ds.vomecrty.data[:,:,dr_w]
        self.data_cross_flow['normal_velocities'].attrs = {'units':'m/s', \
                'standard_name':'contour-normal velocities'}
             
        self.data_cross_flow['normal_transport'] = xr.full_like(u_ds.vozocrtx, np.nan)  
        self.data_cross_flow['normal_transport'][:,:,dr_n] = ( u_ds.vozocrtx.data[:,:,dr_n+1] * 
                                u_ds.e2.data[dr_n+1] * u_ds.e3_0.data[:,dr_n+1] )
        self.data_cross_flow['normal_transport'][:,:,dr_s] = ( -u_ds.vozocrtx.data[:,:,dr_s] * 
                                u_ds.e2.data[dr_s] * u_ds.e3_0.data[:,dr_s] )
        self.data_cross_flow['normal_transport'][:,:,dr_e] = ( -v_ds.vomecrty.data[:,:,dr_e+1] *
                                v_ds.e1.data[dr_e+1] * v_ds.e3_0.data[:,dr_e+1] )
        self.data_cross_flow['normal_transport'][:,:,dr_w] = ( v_ds.vomecrty.data[:,:,dr_w] *
                                v_ds.e1.data[dr_w] * v_ds.e3_0.data[:,dr_w] )
        self.data_cross_flow['normal_transport'].attrs = {'units':'m^3/s', \
                'standard_name':'contour-normal volume transport'}
        
        self.data_cross_flow['depth_integrated_normal_transport'] = (self.data_cross_flow
                                .normal_transport.sum(dim='z_dim') / 1000000.)
        self.data_cross_flow['depth_integrated_normal_transport'].attrs ={'units':'Sv', \
                'standard_name':'contour-normal depth integrated volume transport'}
                                
        self.__update_cross_flow_vars('depth_0',u_ds.depth_0,v_ds.depth_0,dr_n,dr_s,dr_e,dr_w,1)
        self.__update_cross_flow_vars('longitude',u_ds.longitude,v_ds.longitude,dr_n,dr_s,dr_e,dr_w,0)
        self.__update_cross_flow_vars('latitude',u_ds.latitude,v_ds.latitude,dr_n,dr_s,dr_e,dr_w,0)
        self.data_cross_flow['e1'] = xr.full_like(self.data_contour.e1, np.nan)   
        self.__update_cross_flow_vars('e1',u_ds.e1,v_ds.e1,dr_n,dr_s,dr_e,dr_w,0)
        self.data_cross_flow['e2'] = xr.full_like(self.data_contour.e2, np.nan)
        self.__update_cross_flow_vars('e2',u_ds.e2,v_ds.e2,dr_n,dr_s,dr_e,dr_w,0)
        self.data_cross_flow['e3_0'] = xr.full_like(self.data_contour.e3_0, np.nan)
        self.__update_cross_flow_vars('e3_0',u_ds.e3_0,v_ds.e3_0,dr_n,dr_s,dr_e,dr_w,1)
        
        self.data_cross_flow['depth_0'].attrs = {'standard_name':'Depth at time zero \
                on the contour-normal velocity grid points'}
        self.data_cross_flow['latitude'].attrs = {'standard_name':'Latitude at \
                the contour-normal velocity grid points'}
        self.data_cross_flow['longitude'].attrs = {'standard_name':'Longitude at \
                the contour-normal velocity grid points'}

    
    def __update_cross_flow_vars(self, var, u_var,v_var, dr_n, dr_s, dr_e, dr_w, pos ):
        ''' This method will pull variable data at specific points along the contour
        from the u and v grid datasets and put them into the data_cross_flow dataset'''
        if pos==0:
            self.data_cross_flow[var][dr_n] = u_var.data[dr_n+1]
            self.data_cross_flow[var][dr_s] = u_var.data[dr_s]
            self.data_cross_flow[var][dr_e] = v_var.data[dr_e+1]
            self.data_cross_flow[var][dr_w] = v_var.data[dr_w]
            self.data_cross_flow[var][-1] = np.nan  
        elif pos==1:         
            self.data_cross_flow[var][:,dr_n] = u_var.data[:,dr_n+1]
            self.data_cross_flow[var][:,dr_s] = u_var.data[:,dr_s]
            self.data_cross_flow[var][:,dr_e] = v_var.data[:,dr_e+1]
            self.data_cross_flow[var][:,dr_w] = v_var.data[:,dr_w]
            self.data_cross_flow[var][:,-1] = np.nan    
        elif pos==2:         
            self.data_cross_flow[var][:,:,dr_n] = u_var.data[:,:,dr_n+1]
            self.data_cross_flow[var][:,:,dr_s] = u_var.data[:,:,dr_s]
            self.data_cross_flow[var][:,:,dr_e] = v_var.data[:,:,dr_e+1]
            self.data_cross_flow[var][:,:,dr_w] = v_var.data[:,:,dr_w]
            self.data_cross_flow[var][:,:,-1] = np.nan    
        
            
    def __pressure_grad_fpoint2(self, ds_T, ds_T_j1, ds_T_i1, ds_T_j1i1, r_ind, velocity_component):
        """
        Calculates the hydrostatic and surface pressure gradients at a set of f-points
        along the contour, i.e. at a set of specific values of r_dim (but for all time and depth).
        The caller must supply four datasets that contain the variables which define
        the hydrostatic and surface pressure at all vertical z_levels and all time 
        on the t-points around the contour i.e. for a set of f-points on the contour 
        defined each defined at (j+1/2, i+1/2), we want t-points at (j,i), (j+1,i), (j,i+1), (j+1,i+1), 
        corresponding to ds_T, ds_T_j1, ds_T_i1, ds_T_j1i1, respectively. 
        ds_T, ds_T_j1, ds_T_i1, ds_T_j1i1 will have dimensions in time and depth.
        
        The velocity_component defines whether u or v is normal to the contour 
        for the segments of the contour. A segment of contour is 
        defined as being r_dim to r_dim+1 where r_dim is the along contour dimension.


        Returns
        -------
        hpg_f : DataArray with dimensions in time and depth and along contour
            hydrostatic pressure gradient at a set of f-points along the contour
            for all time and depth
        spg_f : DataArray with dimensions in time and depth and along contour
            surface pressure gradient at a set of f-points along the contour

        """
        if velocity_component == "u":
            # required scale factors for derivative and averaging
            e2v = 0.5*( ds_T_j1.e2.data[r_ind] + ds_T.e2.data[r_ind] )
            e2v_i1 = 0.5*( ds_T_j1i1.e2.data[r_ind] + ds_T_i1.e2.data[r_ind] )
            e1v = 0.5*( ds_T_j1.e1.data[r_ind] + ds_T.e1.data[r_ind] )
            e1v_i1 = 0.5*( ds_T_j1i1.e1.data[r_ind] + ds_T_i1.e1.data[r_ind] )
            e1f = 0.5*( e1v + e1v_i1 )            
            # calculate gradients at v-points either side of f-point
            hpg = (ds_T_j1.pressure_h_zlevels.data[:,:,r_ind] - ds_T.pressure_h_zlevels.data[:,:,r_ind]) / e2v
            hpg_i1 = (ds_T_j1i1.pressure_h_zlevels.data[:,:,r_ind] - ds_T_i1.pressure_h_zlevels.data[:,:,r_ind]) / e2v_i1   
            # average onto f-point
            hpg_f = 0.5 * ( ( e1v * hpg ) + ( e1v_i1 * hpg_i1 ) ) / e1f 
            # as aboave            
            spg = (ds_T_j1.pressure_s.data[:,r_ind] - ds_T.pressure_s.data[:,r_ind]) / e2v
            spg_i1 = (ds_T_j1i1.pressure_s.data[:,r_ind] - ds_T_i1.pressure_s.data[:,r_ind]) / e2v_i1
            spg_f = 0.5 * ( (e1v * spg) + (e1v_i1 * spg_i1) ) / e1f 
        elif velocity_component == "v":
            # required scale factors for derivative and averaging
            e1u = 0.5 * ( ds_T_i1.e1.data[r_ind] + ds_T.e1.data[r_ind] ) 
            e1u_j1 = 0.5 * ( ds_T_j1i1.e1.data[r_ind] + ds_T_j1.e1.data[r_ind] )
            e2u = 0.5 * ( ds_T_i1.e2.data[r_ind] + ds_T.e2.data[r_ind] )
            e2u_j1 = 0.5 * ( ds_T_j1i1.e2.data[r_ind] + ds_T_j1.e2.data[r_ind] )
            e2f = 0.5 * ( e2u + e2u_j1 )
            # calculate gradients at u-points either side of f-point
            hpg = (ds_T_i1.pressure_h_zlevels.data[:,:,r_ind] - ds_T.pressure_h_zlevels.data[:,:,r_ind]) / e1u
            hpg_j1 = (ds_T_j1i1.pressure_h_zlevels.data[:,:,r_ind] - ds_T_j1.pressure_h_zlevels.data[:,:,r_ind]) / e1u_j1 
            # average onto f-point
            hpg_f = 0.5 * ( (e2u * hpg) + (e2u_j1 * hpg_j1) ) / e2f
            # as above
            spg = (ds_T_i1.pressure_s.data[:,r_ind] - ds_T.pressure_s.data[:,r_ind]) / e1u
            spg_j1 = (ds_T_j1i1.pressure_s.data[:,r_ind] - ds_T_j1.pressure_s.data[:,r_ind]) / e1u_j1 
            spg_f = 0.5 * ( (e2u * spg) + (e2u_j1 * spg_j1) ) / e2f
        
        return (hpg_f, spg_f)
   
    
    def calc_geostrophic_flow(self, nemo_t: COAsT, ref_density=None):
        """
        This method will calculate the geostrophic velocity and volume transport
        (due to the geostrophic current) across the contour. 
        Four variables are added to the Contour.data_cross_flow dataset:
        1. normal_velocity_hpg      (t_dim, depth_z_levels, r_dim)
        This is the velocity due to the hydrostatic pressure gradient
        2. normal_velocity_spg      (t_dim, r_dim)
        This is the velocity due to the surface pressure gradient
        3. transport_across_AB_hpg  (t_dim, r_dim)
        This is the volume transport due to the hydrostatic pressure gradient
        4. transport_across_AB_spg  (t_dim, r_dim
        This is the volume transport due to the surface pressure gradient
                                                                       
        This implementation works by regridding vertically onto horizontal z_levels in order
        to perform the horizontal gradients. Currently s_level depths are
        assumed fixed at their initial depths, i.e. at time zero.
        
        Requirements: The nemo t-grid dataset, nemo_t, must contain the sea surface height,
        Practical Salinity and the Potential Temperature variables. The depth_0
        field must also be supplied. The GSW package is used to calculate
        The Absolute Pressure, Absolute Salinity and Conservate Temperature.
        
        Parameters
        ----------
        nemo_t : COAsT
            This is the nemo model data on the t-grid for the entire domain.
        ref_density : TYPE, optional
            reference density value. If not supplied a mean in time, depth and 
            along the contour will be used as the mean reference value.

        Returns
        -------
        None.

        """
        # If there is no time dimension, add one then remove at end. This is so
        # indexing can assume a time dimension exists
        nemo_t_local = nemo_t.copy()
        if 't_dim' not in nemo_t_local.dataset.dims:
            nemo_t_local.dataset = nemo_t_local.dataset.expand_dims(dim={'t_dim':1},axis=0)
        
        # We need to calculate the pressure at four t-points to get an
        # average onto the pressure gradient at the f-points, which will then
        # be averaged onto the normal velocity points. Here we subset the nemo_t 
        # data around the contour so we have these four t-grid points at each 
        # point along the contour        
        cont_t = Contour_t(nemo_t_local, self.y_ind, self.x_ind, self.depth)            # j,i
        cont_t_j1 = Contour_t(nemo_t_local, self.y_ind+1, self.x_ind, self.depth)       # j+1,i
        cont_t_i1 = Contour_t(nemo_t_local, self.y_ind, self.x_ind+1, self.depth)       # j,i+1
        cont_t_j1i1 = Contour_t(nemo_t_local, self.y_ind+1, self.x_ind+1, self.depth)   # j+1,i+1
        
        bath_max = np.max([cont_t.data_contour.bathymetry.max().item(), 
                           cont_t_j1.data_contour.bathymetry.max().item(),
                           cont_t_i1.data_contour.bathymetry.max().item(), 
                           cont_t_j1i1.data_contour.bathymetry.max().item()])   
        z_levels = self.gen_z_levels(bath_max)
        
        cont_t.construct_pressure(ref_density, z_levels, extrapolate=True)
        cont_t_j1.construct_pressure(ref_density, z_levels, extrapolate=True)
        cont_t_i1.construct_pressure(ref_density, z_levels, extrapolate=True)
        cont_t_j1i1.construct_pressure(ref_density, z_levels, extrapolate=True)        
        
        # Remove the mean hydrostatic pressure on each z_level from the hydrostatic pressure.
        # This helps to reduce the noise when taking the horizontal gradients of hydrostatic pressure.
        # Also catch and ignore nan-slice warning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            pressure_h_zlevel_mean = ( xr.concat( (cont_t.data_contour.pressure_h_zlevels, 
                            cont_t_j1.data_contour.pressure_h_zlevels, 
                            cont_t_i1.data_contour.pressure_h_zlevels, 
                            cont_t_j1i1.data_contour.pressure_h_zlevels), dim='concat_dim' )
                            .mean(dim=('concat_dim','r_dim','t_dim'),skipna=True) )
            if ref_density is None:
                ref_density = ( xr.concat( (cont_t.data_contour.density_zlevels, 
                            cont_t_j1.data_contour.density_zlevels, 
                            cont_t_i1.data_contour.density_zlevels, 
                            cont_t_j1i1.data_contour.density_zlevels), dim='concat_dim' )
                            .mean(dim=('concat_dim','r_dim','t_dim','depth_z_levels'),skipna=True).item() )
        cont_t.data_contour['pressure_h_zlevels'] = \
                cont_t.data_contour.pressure_h_zlevels - pressure_h_zlevel_mean
        cont_t_j1.data_contour['pressure_h_zlevels'] = \
                cont_t_j1.data_contour.pressure_h_zlevels - pressure_h_zlevel_mean
        cont_t_i1.data_contour['pressure_h_zlevels'] = \
                cont_t_i1.data_contour.pressure_h_zlevels - pressure_h_zlevel_mean
        cont_t_j1i1.data_contour['pressure_h_zlevels'] = \
                cont_t_j1i1.data_contour.pressure_h_zlevels - pressure_h_zlevel_mean
                
        # Coriolis parameter
        f = 2 * self.EARTH_ROT_RATE * np.sin( np.deg2rad(self.data_contour.latitude) )        
    
        # Find the indices where the derivative of the contour in the north, south, east and west
        # directions are positive.
        dr_n = np.where(np.diff(self.y_ind)>0, np.arange(0,self.data_contour.r_dim.size-1), np.nan )
        dr_s = np.where(np.diff(self.y_ind)<0, np.arange(0,self.data_contour.r_dim.size-1), np.nan )
        dr_e = np.where(np.diff(self.x_ind)>0, np.arange(0,self.data_contour.r_dim.size-1), np.nan )
        dr_w = np.where(np.diff(self.x_ind)<0, np.arange(0,self.data_contour.r_dim.size-1), np.nan )        
        dr_list = [dr_n[~np.isnan(dr_n)].astype(int), dr_s[~np.isnan(dr_s)].astype(int),
                   dr_e[~np.isnan(dr_e)].astype(int), dr_w[~np.isnan(dr_w)].astype(int)]
        
        # horizontal scale factors on the relevent u and v grids that are
        # normal to the contour for dr_n, dr_s, dr_e, dr_w
        e2u_j1  = 0.5 * ( cont_t_j1.data_contour.e2.data[dr_list[0]] + cont_t_j1i1.data_contour.e2.data[dr_list[0]] )
        e2u     = 0.5 * ( cont_t.data_contour.e2.data[dr_list[1]] + cont_t_i1.data_contour.e2.data[dr_list[1]] )
        e1v_i1  = 0.5 * ( cont_t_i1.data_contour.e1.data[dr_list[2]] + cont_t_j1i1.data_contour.e1.data[dr_list[2]] )
        e1v     = 0.5 * ( cont_t.data_contour.e1.data[dr_list[3]] + cont_t_j1.data_contour.e1.data[dr_list[3]] )
        e_horiz_vel = [e2u_j1, e2u, e1v_i1, e1v] 
        # Horizontal scale factors on f-grid for dr_n, dr_s, dr_e, dr_w
        e_horiz_f   = [self.data_contour.e2, self.data_contour.e2, 
                       self.data_contour.e1, self.data_contour.e1]
        # velocity component normal to contour for dr_n, dr_s, dr_e, dr_w
        velocity_component = ["u","u","v",'v']
        # Geostrophic flow direction across contour
        flow_direction = [-1,1,-1,1]        
        
        normal_velocity_hpg = np.zeros_like(cont_t.data_contour.pressure_h_zlevels)
        normal_velocity_spg = np.zeros_like(cont_t.data_contour.pressure_s)
        # horizontal scale factors for each segmant of contour
        e_horiz = np.zeros( (cont_t.data_contour.t_dim.size, cont_t.data_contour.r_dim.size) ) 
        # Contruct geostrophic flow
        for dr, vel_comp, flow_dir, e_hor_vel, e_hor_f in \
                zip(dr_list, velocity_component, flow_direction, e_horiz_vel, e_horiz_f) :
            hpg, spg        = self.__pressure_grad_fpoint2( cont_t.data_contour, 
                                cont_t_j1.data_contour, cont_t_i1.data_contour, 
                                cont_t_j1i1.data_contour, dr, vel_comp )
            hpg_r1, spg_r1  = self.__pressure_grad_fpoint2( cont_t.data_contour,
                                cont_t_j1.data_contour,
                                cont_t_i1.data_contour, 
                                cont_t_j1i1.data_contour, dr+1, vel_comp )
            normal_velocity_hpg[:,:,dr] = ( flow_dir * 0.5 * (e_hor_f.data[dr]*hpg/f.data[dr] 
                                            + e_hor_f.data[dr+1]*hpg_r1/f.data[dr+1]) 
                                            / (e_hor_vel * ref_density) )
            normal_velocity_spg[:,dr]   = ( flow_dir * 0.5 * (e_hor_f.data[dr]*spg/f.data[dr] 
                                            + e_hor_f.data[dr+1]*spg_r1/f.data[dr+1]) 
                                            / (e_hor_vel * ref_density) ) 
            e_horiz[:,dr]               = e_hor_vel

        # Bathymetry at normal velocity points             
        H = np.zeros_like( self.data_contour.bathymetry.values )
        H[:-1] = 0.5*(self.data_contour.bathymetry.values[:-1] + self.data_contour.bathymetry.values[1:])
        # Remove redundent levels below bathymetry
        normal_velocity_hpg = np.where( z_levels[:,np.newaxis] <= H, 
                                       normal_velocity_hpg, np.nan )           
        active_z_levels = np.count_nonzero(~np.isnan(normal_velocity_hpg),axis=1).max() 
        normal_velocity_hpg = normal_velocity_hpg[:,:active_z_levels,:]
        z_levels = z_levels[:active_z_levels]

        # The cross contour flow is defined on the u and v points that are across
        # the contour, i.e. between f points, therefore the attributes of the
        # data_cross_flow dataset need to be on these points.
        da_y_ind = xr.DataArray( self.y_ind, dims=['r_dim'] )
        da_x_ind = xr.DataArray( self.x_ind, dims=['r_dim'] )
        u_ds = NEMO( fn_domain=self.filename_domain, grid_ref='u-grid' ).dataset \
            .isel(y_dim=da_y_ind, x_dim=da_x_ind)
        v_ds = NEMO( fn_domain=self.filename_domain, grid_ref='v-grid' ).dataset \
            .isel(y_dim=da_y_ind, x_dim=da_x_ind)
        self.data_cross_flow['e1'] = xr.full_like(self.data_contour.e1, np.nan)  
        self.data_cross_flow['e2'] = xr.full_like(self.data_contour.e2, np.nan)
        self.data_cross_flow['latitude'] = xr.full_like(self.data_contour.latitude, np.nan)  
        self.data_cross_flow['longitude'] = xr.full_like(self.data_contour.longitude, np.nan)
        self.__update_cross_flow_vars('longitude',u_ds.longitude,v_ds.longitude,
                            dr_list[0],dr_list[1],dr_list[2],dr_list[3],0)
        self.__update_cross_flow_vars('latitude',u_ds.latitude,v_ds.latitude,
                            dr_list[0],dr_list[1],dr_list[2],dr_list[3],0)         
        self.__update_cross_flow_vars('e1',u_ds.e1,v_ds.e1,
                            dr_list[0],dr_list[1],dr_list[2],dr_list[3],0)        
        self.__update_cross_flow_vars('e2',u_ds.e2,v_ds.e2,
                            dr_list[0],dr_list[1],dr_list[2],dr_list[3],0)
        self.data_cross_flow['latitude'].attrs = {'standard_name':'Latitude at \
                the contour-normal velocity grid points'}
        self.data_cross_flow['longitude'].attrs = {'standard_name':'Longitude at \
                the contour-normal velocity grid points'}
        
        # DataArray attributes
        coords_hpg={'depth_z_levels': (('depth_z_levels'), z_levels),
                'latitude': (('r_dim'), self.data_cross_flow.latitude),
                'longitude': (('r_dim'), self.data_cross_flow.longitude)}
        dims_hpg=['depth_z_levels', 'r_dim']
        attributes_hpg = {'units': 'm/s', 'standard name': 'velocity across the \
                          transect due to the hydrostatic pressure gradient'}
        coords_spg={'latitude': (('r_dim'), self.data_cross_flow.latitude),
                'longitude': (('r_dim'), self.data_cross_flow.longitude)}
        dims_spg=['r_dim']
        attributes_spg = {'units': 'm/s', 'standard name': 'velocity across the \
                          transect due to the surface pressure gradient'}
        
        # Add time if required
        if 't_dim' in cont_t.data_contour.dims:
            coords_hpg['time'] = (('t_dim'), cont_t.data_contour.time)
            dims_hpg.insert(0, 't_dim')
            coords_spg['time'] = (('t_dim'), cont_t.data_contour.time)
            dims_spg.insert(0, 't_dim')
        
        # Add DataArrays  to dataset
        self.data_cross_flow['normal_velocity_hpg'] = xr.DataArray( np.squeeze(normal_velocity_hpg),
                coords=coords_hpg, dims=dims_hpg, attrs=attributes_hpg)
        self.data_cross_flow['normal_velocity_spg'] = xr.DataArray( np.squeeze(normal_velocity_spg),
                coords=coords_spg, dims=dims_spg, attrs=attributes_spg)
        self.data_cross_flow['transport_across_AB_hpg'] = ( self.data_cross_flow
                .normal_velocity_hpg.fillna(0).integrate(dim='depth_z_levels') ) * e_horiz / 1000000   
        self.data_cross_flow.transport_across_AB_hpg.attrs = {'units': 'Sv', 
                'standard_name': 'volume transport across transect due to the hydrostatic pressure gradient'}        
        self.data_cross_flow['transport_across_AB_spg'] = self.data_cross_flow.normal_velocity_spg * H * e_horiz / 1000000
        self.data_cross_flow.transport_across_AB_spg.attrs = {'units': 'Sv', 
                'standard_name': 'volume transport across transect due to the surface pressure gradient'}
        
        
class Contour_t(Contour):
    '''
    Class defining a Contour type on the t-grid, which is a 3d dataset of points between a point A and 
    a point B defining an isobath contour. The dataset has a time, depth and contour dimension. 
    The contour dimension defines the points along the contour.
    The supplied model t-grid Data is subsetted in its entirety along these dimensions and
    calculations can be performed on this dataset.
    
    Parameters
    ----------
    nemo : COAsT
        t-grid nemo object containing the model dataset.
    y_ind : numpy.ndarray
        1d array of y indices defining the contour on the model grid
    x_ind : numpy.ndarray
        1d array of x indices defining the contour on the model grid
    depth : int
        Depth of contour isobath
    '''
    
    def __init__(self, nemo_t: COAsT, y_ind, x_ind, depth):
        super().__init__(nemo_t, y_ind, x_ind, depth)
        
        
    def construct_pressure( self, ref_density=None, z_levels=None, extrapolate=False ):   
        '''
            This method is for calculating the hydrostatic and surface pressure fields
            on horizontal levels in the vertical (z-levels). The motivation 
            is to enable the calculation of horizontal gradients; however, 
            the variables can quite easily be interpolated onto the original 
            vertical grid.
             
            Requirements: The object's t-grid dataset must contain the sea surface height,
            Practical Salinity and the Potential Temperature variables.
            The GSW package is used to calculate the Absolute Pressure, 
            Absolute Salinity and Conservate Temperature.
            
            Three new variables (density, hydrostatic pressure, surface pressure)
            are created and added to the Contour_t.data_contour dataset:
                density_zlevels       (t_dim, depth_z_levels, r_dim)
                pressure_h_zlevels    (t_dim, depth_z_levels, r_dim)
                pressure_s            (t_dim, r_dim)
            
            Note that density is constructed using the EOS10
            equation of state.

        Parameters
        ----------
        ref_density: float
            reference density value, if None, then the Contour mean across time, 
            depth and along contour will be used.
        z_levels : (optional) numpy array
            1d array that defines the depths to interpolate the density and pressure
            on to.
        extrapolate : boolean, default False
            If true the variables are extrapolated to the deepest z_level, if false,
            values below the bathymetry are set to NaN
        Returns
        -------
        None.

        '''        
        
        # If there is no time dimension, add one, this is so
        # indexing can assume a time dimension exists
        if 't_dim' not in self.data_contour.dims:
            self.data_contour = self.data_contour.expand_dims(dim={'t_dim':1},axis=0)

        # Generate vertical levels if not supplied
        if z_levels is None:   
            z_levels = self.gen_z_levels( self.data_contour.bathymetry.max().item() )
        
        shape_ds = ( self.data_contour.t_dim.size, len(z_levels), self.data_contour.r_dim.size )
        salinity_z = np.ma.zeros( shape_ds )
        temperature_z = np.ma.zeros( shape_ds ) 
        salinity_s = self.data_contour.salinity.to_masked_array()
        temperature_s = self.data_contour.temperature.to_masked_array()
        s_levels = self.data_contour.depth_0.values
        
        # Interpolate salinity and temperature onto z-levels
        # Note. At the current time there does not appear to be a good algorithm for 
        # performing this type of interpolation without loops, which can be a bottleneck.
        # Griddata is an option but does not support extrapolation and did not 
        # have noticable performance benefit.
        for it in self.data_contour.t_dim:
            for ir in self.data_contour.r_dim:
                if not np.all(np.isnan(salinity_s[it,:,ir].data)):  
                    # Need to remove the levels below the (envelope) bathymetry which are NaN
                    salinity_s_r = salinity_s[it,:,ir].compressed()
                    temperature_s_r = temperature_s[it,:,ir].compressed()
                    s_levels_r = s_levels[:len(salinity_s_r),ir]
                    
                    sal_func = interpolate.interp1d( s_levels_r, salinity_s_r, 
                                 kind='linear', fill_value="extrapolate")
                    temp_func = interpolate.interp1d( s_levels_r, temperature_s_r, 
                                 kind='linear', fill_value="extrapolate")
                    
                    if extrapolate is True:
                        salinity_z[it,:,ir] = sal_func(z_levels)
                        temperature_z[it,:,ir] = temp_func(z_levels)                        
                    else:
                        # set levels below the bathymetry to nan
                        salinity_z[it,:,ir] = np.where( z_levels <= self.data_contour.bathymetry.values[ir], 
                                sal_func(z_levels), np.nan )
                        temperature_z[it,:,ir] = np.where( z_levels <= self.data_contour.bathymetry.values[ir], 
                                temp_func(z_levels), np.nan ) 
                    
        if extrapolate is False:
            # remove redundent levels    
            active_z_levels = np.count_nonzero(~np.isnan(salinity_z),axis=1).max() 
            salinity_z = salinity_z[:,:active_z_levels,:]
            temperature_z = temperature_z[:,:active_z_levels,:]
            z_levels = z_levels[:active_z_levels]
        
        # Absolute Pressure (depth must be negative)   
        pressure_absolute = np.ma.masked_invalid(
            gsw.p_from_z( -z_levels[:,np.newaxis], self.data_contour.latitude ) )         
        # Absolute Salinity           
        salinity_absolute = np.ma.masked_invalid(
            gsw.SA_from_SP( salinity_z, pressure_absolute, self.data_contour.longitude, 
            self.data_contour.latitude ) )
        salinity_absolute = np.ma.masked_less(salinity_absolute,0)
        # Conservative Temperature
        temp_conservative = np.ma.masked_invalid(
            gsw.CT_from_pt( salinity_absolute, temperature_z ) )
        # In-situ density
        density_z = np.ma.masked_invalid( gsw.rho( 
            salinity_absolute, temp_conservative, pressure_absolute ) )
        
        coords={'depth_z_levels': (('depth_z_levels'), z_levels),
                'latitude': (('r_dim'), self.data_contour.latitude),
                'longitude': (('r_dim'), self.data_contour.longitude)}
        dims=['depth_z_levels', 'r_dim']
        attributes = {'units': 'kg / m^3', 'standard name': 'In-situ density on the z-level vertical grid'}
        
        if shape_ds[0] != 1:
            coords['time'] = (('t_dim'), self.data_contour.time.values)
            dims.insert(0, 't_dim')
         
        if ref_density is None:    
            ref_density = np.mean( density_z )
        self.data_contour['density_zlevels'] = xr.DataArray( np.squeeze(density_z), 
                coords=coords, dims=dims, attrs=attributes )
    
        # Cumulative integral of perturbation density on z levels
        density_cumulative = -cumtrapz( density_z - ref_density, x=-z_levels, axis=1, initial=0)
        hydrostatic_pressure = density_cumulative * self.GRAVITY
        
        attributes = {'units': 'kg m^{-1} s^{-2}', 'standard name': 'Hydrostatic perturbation pressure on the z-level vertical grid'}
        self.data_contour['pressure_h_zlevels'] = xr.DataArray( np.squeeze(hydrostatic_pressure), 
                coords=coords, dims=dims, attrs=attributes )        
        self.data_contour['pressure_s'] = ref_density * self.GRAVITY * self.data_contour.ssh.squeeze()
        self.data_contour.pressure_s.attrs = {'units': 'kg m^{-1} s^{-2}', 
                                  'standard_name': 'Surface perturbation pressure'}
        
    
        
        
    


from .COAsT import COAsT
from scipy.ndimage import convolve1d
from scipy import interpolate
import gsw
import xarray as xr
import numpy as np
import math
from scipy.interpolate import griddata
from scipy.integrate import cumtrapz, trapz
from .logging_util import get_slug, debug, error
import warnings

# =============================================================================
# The TRANSECT module is a place for code related to transects only
# =============================================================================

class Transect:
    GRAVITY = 9.8 
    
    def __init__(self, point_A: tuple, point_B: tuple, nemo_F: COAsT,
                 nemo_T: COAsT=None, nemo_U: COAsT=None, nemo_V: COAsT=None):
        '''
        Class defining a generic transect type, which is a 3d dataset between a point A and 
        a point B, with a time dimension, a depth dimension and a transect dimension. The 
        transect dimension defines the points along the transect.
        The model Data is subsetted in its entirety along these dimensions.
        
        Note that Point A should be of lower latitude than point B.
        
        Example usage:
            point_A = (54,-15)
            point_B = (56,-12)
            transect = coast.Transect( point_A, point_B, nemo_t, nemo_u, nemo_v, nemo_f )
            

        Parameters
        ----------
        point_A : tuple, (lat,lon)
        point_B : tuple, (lat,lon)
        nemo_F : NEMO, model data on the F grid, which is required to define the 
                    transect
        nemo_T : NEMO, optional, model data on the T grid
        nemo_U : NEMO, optional, model data on the U grid
        nemo_V : NEMO, optional, model data on the V grid
        
        
        '''
        debug(f"Creating a new {get_slug(self)}")
        # point A should be of lower latitude than point B
        if abs(point_B[0]) < abs(point_A[0]):
            self.point_A = point_B
            self.point_B = point_A
        else:
            self.point_A = point_A
            self.point_B = point_B
            
        # self.nemo_F = nemo_F
        # self.nemo_U = nemo_U
        # self.nemo_V = nemo_V
        # self.nemo_T = nemo_T
        self.filename_domain = nemo_F.filename_domain
            
        # Get points on transect
        tran_y_ind, tran_x_ind = self.get_transect_indices( nemo_F )
                
        # indices along the transect        
        self.y_ind = tran_y_ind
        self.x_ind = tran_x_ind
        self.len = len(tran_y_ind)
        self.data_tran = xr.Dataset()
        
        # Subset the nemo data along the transect creating a new dimension (r_dim),
        # which is a paramterisation for x_dim and y_dim defining the transect
        da_tran_y_ind = xr.DataArray( tran_y_ind, dims=['r_dim'])
        da_tran_x_ind = xr.DataArray( tran_x_ind, dims=['r_dim'])
        self.data_F = nemo_F.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        if nemo_T is not None:
            self.data_T = nemo_T.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        if nemo_U is not None:
            self.data_U = nemo_U.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        if nemo_V is not None:
            self.data_V = nemo_V.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        # For calculations we need access to a halo of points around the transect
        # self.data_n = dataset.isel(y=tran_y+1,x=tran_x)  
        # self.data_e = dataset.isel(y=tran_y,x=tran_x+1) 
        # self.data_s = dataset.isel(y=tran_y-1,x=tran_x) 
        # self.data_w = dataset.isel(y=tran_y,x=tran_x-1)
        debug(f"{get_slug(self)} initialised")

    def get_transect_indices(self, nemo_F):
        '''
        Get the transect indices on a specific grid

        Parameters
        ----------
        nemo_F : the model grid to define the transect on
        
        Return
        ----------
        tran_y : array of y_dim indices
        tran_x : array of x_dim indices

        '''
        debug(f"Fetching transect indices for {get_slug(self)} with {get_slug(nemo_F)}")
        tran_y_ind, tran_x_ind, tran_len = nemo_F.transect_indices(self.point_A, self.point_B)
        tran_y_ind = np.asarray(tran_y_ind)
        tran_x_ind = np.asarray(tran_x_ind)
        
        # Redefine transect so that each point on the transect is seperated
        # from its neighbours by a single index change in y or x, but not both
        dist_option_1 = nemo_F.dataset.e2.values[tran_y_ind, tran_x_ind] + nemo_F.dataset.e1.values[tran_y_ind+1, tran_x_ind]
        dist_option_2 = nemo_F.dataset.e2.values[tran_y_ind, tran_x_ind+1] + nemo_F.dataset.e1.values[tran_y_ind, tran_x_ind]
        spacing = np.abs( np.diff(tran_y_ind) ) + np.abs( np.diff(tran_x_ind) )
        spacing[spacing!=2]=0
        doublespacing = np.nonzero( spacing )[0]
        for ispacing in doublespacing[::-1]:
            if dist_option_1[ispacing] < dist_option_2[ispacing]:
                tran_y_ind = np.insert( tran_y_ind, ispacing+1, tran_y_ind[ispacing+1] )
                tran_x_ind = np.insert( tran_x_ind, ispacing+1, tran_x_ind[ispacing] )
            else:
                tran_y_ind = np.insert( tran_y_ind, ispacing+1, tran_y_ind[ispacing] )
                tran_x_ind = np.insert( tran_x_ind, ispacing+1, tran_x_ind[ispacing+1] ) 
        return (tran_y_ind, tran_x_ind)
    
    def transport_across_AB(self):
        """
    
        Computes the flow through the transect at each segment and stores:
        Transect normal velocities at each grid point in Transect.normal_velocities,
        Depth integrated volume transport across the transect at each transect segment in 
        Transect.depth_integrated_transport_across_AB
        
        Return 
        -----------
        Transect normal velocities at each grid point (m/s)
        Depth integrated volume transport across the transect at each transect segment (Sv)
        """
        debug(f"Computing transport across AB for {get_slug(self)}")  # TODO Probably want a better description here
        velocity = np.ma.zeros(np.shape(self.data_U.vozocrtx))
        vol_transport = np.ma.zeros(np.shape(self.data_U.vozocrtx))
        depth_integrated_transport = np.ma.zeros( np.shape(self.data_U.vozocrtx[:,0,:] )) 
        depth_0 = np.ma.zeros(np.shape(self.data_U.depth_0))
        
        dy = np.diff(self.y_ind)
        dx = np.diff(self.x_ind)
        # Loop through each point along the transact
        for idx in np.arange(0, self.len-1):            
            if dy[idx] > 0:
                # u flux (+ in)
                velocity[:,:,idx] = self.data_U.vozocrtx[:,:,idx+1].to_masked_array()
                vol_transport[:,:,idx] = ( velocity[:,:,idx] * self.data_U.e2[idx+1].to_masked_array() *
                                          self.data_U.e3_0[:,idx+1].to_masked_array() )
                depth_integrated_transport[:,idx] = np.sum( vol_transport[:,:,idx], axis=1 )
                depth_0[:,idx] = self.data_U.depth_0[:,idx+1].to_masked_array()
            elif dx[idx] > 0:
                # v flux (- in) 
                velocity[:,:,idx] = - self.data_V.vomecrty[:,:,idx+1].to_masked_array()
                vol_transport[:,:,idx] = ( velocity[:,:,idx] * self.data_V.e1[idx+1].to_masked_array() *
                                          self.data_V.e3_0[:,idx+1].to_masked_array() )
                depth_integrated_transport[:,idx] = np.sum( vol_transport[:,:,idx], axis=1 )
                depth_0[:,idx] = self.data_V.depth_0[:,idx+1].to_masked_array()
            elif dx[idx] < 0:
                # v flux (+ in)
                velocity[:,:,idx] = self.data_V.vomecrty[:,:,idx].to_masked_array()
                vol_transport[:,:,idx] = ( velocity[:,:,idx] * self.data_V.e1[idx].to_masked_array() *
                                          self.data_V.e3_0[:,idx].to_masked_array() )
                depth_integrated_transport[:,idx] = np.sum( vol_transport[:,:,idx], axis=1 )
                depth_0[:,idx] = self.data_V.depth_0[:,idx].to_masked_array()
        
        #dimensions = ['t_dim', 'z_dim', 'r_dim']
        
        velocity[:,:,-1] = np.nan
        depth_integrated_transport[:,-1] = np.nan
        
        self.data_tran['normal_velocities'] = xr.DataArray( velocity, 
                    coords={'time': (('t_dim'), self.data_U.time.values),'depth_0': (('z_dim','r_dim'), depth_0)},
                    dims=['t_dim', 'z_dim', 'r_dim'] )        
    
        self.data_tran['depth_integrated_transport_across_AB'] = xr.DataArray( depth_integrated_transport / 1000000.,
                    coords={'time': (('t_dim'), self.data_U.time.values)},
                    dims=['t_dim', 'r_dim'] ) 
        
        self.data_tran.depth_0.attrs['units'] = 'm'
        self.data_tran.depth_0.attrs['standard_name'] = 'Initial depth at time zero'
        self.data_tran.depth_0.attrs['long_name'] = 'Initial depth at time zero defined at the normal velocity grid points on the transect'
                        
        return  # TODO Should this return something? If not then the statement is not needed

    def __pressure_grad_fpoint(self, ds_T, ds_T_j1, ds_T_i1, ds_T_j1i1, velocity_component):
        """
        Calculates the hydrostatic and surface pressure gradients at an f-point
        along the transect, i.e. a specific value of r_dim (but for all time and depth).
        The caller must supply four datasets that define
        the hydrostatic and surface pressure at all vertical z_levels and all time 
        on the t-points around the transect i.e. for an f-point on the transect 
        defined at (j+1/2, i+1/2), we want t-points at (j,i), (j+1,i), (j,i+1), (j+1,i+1), 
        corresponding to  ds_T, ds_T_j1, ds_T_i1, ds_T_j1i1, respectively. 
        ds_T, ds_T_j1, ds_T_i1, ds_T_j1i1 will have dimensions in time and depth.
        
        The velocity_component defines whether u or v is normal to the transect 
        for that particular segment of the transect. A segment of transect is 
        defined as being r_dim to r_dim+1 where r_dim is the dimension along the transect.


        Returns
        -------
        hpg_f : DataArray with dimensions in time and depth
            hydrostatic pressure gradient at an f-point point along the transect
            for all time and depth
        spg_f : DataArray with dimensions in time and depth
            surface pressure gradient at an f-point point along the transect

        """
        if velocity_component == "u":
            # required scale factors for derivative and averaging
            e2v = 0.5*( ds_T_j1.e2 + ds_T.e2 )
            e2v_i1 = 0.5*( ds_T_j1i1.e2 + ds_T_i1.e2 )
            e1v = 0.5*( ds_T_j1.e1 + ds_T.e1 )
            e1v_i1 = 0.5*( ds_T_j1i1.e1 + ds_T_i1.e1 )
            e1f = 0.5*( e1v + e1v_i1 )            
            # calculate gradients at v-points either side of f-point
            hpg = (ds_T_j1.pressure_h_zlevels - ds_T.pressure_h_zlevels) / e2v
            hpg_i1 = (ds_T_j1i1.pressure_h_zlevels - ds_T_i1.pressure_h_zlevels) / e2v_i1   
            # average onto f-point
            hpg_f = 0.5 * ( ( e1v * hpg ) + ( e1v_i1 * hpg_i1 ) ) / e1f 
            # as aboave            
            spg = (ds_T_j1.pressure_s - ds_T.pressure_s) / e2v
            spg_i1 = (ds_T_j1i1.pressure_s - ds_T_i1.pressure_s) / e2v_i1
            spg_f = 0.5 * ( (e1v * spg) + (e1v_i1 * spg_i1) ) / e1f 
        elif velocity_component == "v":
            # required scale factors for derivative and averaging
            e1u = 0.5 * ( ds_T_i1.e1 + ds_T.e1 ) 
            e1u_j1 = 0.5 * ( ds_T_j1i1.e1 + ds_T_j1.e1 )
            e2u = 0.5 * ( ds_T_i1.e2 + ds_T.e2 )
            e2u_j1 = 0.5 * ( ds_T_j1i1.e2 + ds_T_j1.e2 )
            e2f = 0.5 * ( e2u + e2u_j1 )
            # calculate gradients at u-points either side of f-point
            hpg = (ds_T_i1.pressure_h_zlevels - ds_T.pressure_h_zlevels) / e1u
            hpg_j1 = (ds_T_j1i1.pressure_h_zlevels - ds_T_j1.pressure_h_zlevels) / e1u_j1 
            # average onto f-point
            hpg_f = 0.5 * ( (e2u * hpg) + (e2u_j1 * hpg_j1) ) / e2f
            # as above
            spg = (ds_T_i1.pressure_s - ds_T.pressure_s) / e1u
            spg_j1 = (ds_T_j1i1.pressure_s - ds_T_j1.pressure_s) / e1u_j1 
            spg_f = 0.5 * ( (e2u * spg) + (e2u_j1 * spg_j1) ) / e2f
        
        return (hpg_f, spg_f)
    

   
    def geostrophic_transport(self, nemo_t_object: COAsT, ref_density = 1027):
        """
        This method will calculate the geostrophic velocity and volume transport
        (due to the geostrophic vurrent) across the transect. 
        4 variables are added to the TRANSECT.tran_data dataset:
            1. normal_velocity_hpg      (t_dim, depth_z_levels, r_dim)
            This is the velocity due to the hydrostatic pressure gradient
            2. normal_velocity_spg      (t_dim, r_dim)
            This is the velocity due to the surface pressure gradient
            3. transport_across_AB_hpg  (t_dim, r_dim)
            This is the volume transport due to the hydrostatic pressure gradient
            4. transport_across_AB_spg  (t_dim, r_dim
            This is the volume transport due to the surface pressure gradient
                                                                       
        Ths implementation works by regridding from s_levels to z_levels in order
        to perform the horizontal gradients. Currently the s_level depths are
        assumed fixed at their initial depths, i.e. at time zero.
        
        
        
        Parameters
        ----------
        nemo_t_object : COAsT
            This is the nemo model data on the t-grid for the entire domain. It
            must contain the temperature, salinity and t-grid domain data (e1t, e2t, e3t_0).
        ref_density : TYPE, optional
            reference density value. The default is 1027.

        Returns
        -------
        None.

        """
        debug(f"Calculating geostrophic velocity and volume transport for {get_slug(self)} with "
              f"{get_slug(nemo_t_object)}")
        nemo_T = nemo_t_object.copy()
        if 't_dim' not in nemo_T.dataset.dims:
            nemo_T_ds = nemo_T.dataset.expand_dims(dim={'t_dim':1},axis=0)
        else:
            nemo_T_ds = nemo_T.dataset 
        
        nemo_T_ds['bathymetry'] = xr.open_dataset( self.filename_domain ).bathy_metry.squeeze().rename({'y':'y_dim', 'x':'x_dim'}) 
               
        y_ind = xr.DataArray( self.y_ind, dims=['r_dim'] ) # j
        x_ind = xr.DataArray( self.x_ind, dims=['r_dim'] ) # i
        
        # We need to calculate the pressure at four t-points to average onto the
        # normal velocity points. Here we subset the nemo_t data around the
        # transect so we have these four t-grid points at each point along the 
        # transect
        ds_T = nemo_T_ds.isel(y_dim = y_ind, x_dim = x_ind) # j,i
        ds_T_j1 = nemo_T_ds.isel(y_dim = y_ind+1, x_dim = x_ind) # j+1,i
        ds_T_i1 = nemo_T_ds.isel(y_dim = y_ind, x_dim = x_ind+1) # j,i+1
        ds_T_j1i1 = nemo_T_ds.isel(y_dim = y_ind+1, x_dim = x_ind+1) # j+1,i+1
        
        # Construct and add the pressure fields to the 4 t-grid datasets
        self.construct_pressure(ds_T, ref_density)
        self.construct_pressure(ds_T_j1, ref_density)
        self.construct_pressure(ds_T_i1, ref_density)
        self.construct_pressure(ds_T_j1i1, ref_density)
        
        # Remove the mean hydrostatic pressure on each z_level from the hydrostatic pressure.
        # This helps to reduce the noise when taking the horizontal gradients of hydrostatic pressure.
        # Also catch and ignore nan-slice warning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            pressure_h_zlevel_mean = xr.concat( (ds_T.pressure_h_zlevels, ds_T_j1.pressure_h_zlevels, 
                                 ds_T_i1.pressure_h_zlevels, ds_T_j1i1.pressure_h_zlevels), 
                                 dim='concat_dim' ).mean(dim=('concat_dim','r_dim','t_dim'),skipna=True)
        ds_T['pressure_h_zlevels'] = ds_T.pressure_h_zlevels - pressure_h_zlevel_mean
        ds_T_j1['pressure_h_zlevels'] = ds_T_j1.pressure_h_zlevels - pressure_h_zlevel_mean
        ds_T_i1['pressure_h_zlevels'] = ds_T_i1.pressure_h_zlevels - pressure_h_zlevel_mean
        ds_T_j1i1['pressure_h_zlevels'] = ds_T_j1i1.pressure_h_zlevels - pressure_h_zlevel_mean
                
        # Coriolis parameter
        f = 2 * 7.2921 * 10**(-5) * np.sin( np.deg2rad(self.data_F.latitude) )
        
        dy = np.diff(self.y_ind)
        dx = np.diff(self.x_ind)
        normal_velocity_hpg = np.zeros_like(ds_T.pressure_h_zlevels)
        normal_velocity_spg = np.zeros_like(ds_T.pressure_s)
        # horizontal scale factors for each segmant of transect
        horizontal_scale = np.zeros( (ds_T.t_dim.size, ds_T.r_dim.size) ) 
        # Loop through each point along the transact
        for idx in np.arange(0, self.len-1):  
            # u flux (+u is positive across transect) 
            if dy[idx] > 0:    
                # calculate the pressure gradients at two f points defining a segment of the transect                             
                hpg, spg = self.__pressure_grad_fpoint( ds_T.isel(r_dim=idx), ds_T_j1.isel(r_dim=idx),
                                        ds_T_i1.isel(r_dim=idx), ds_T_j1i1.isel(r_dim=idx), 'u' )
                hpg_r1, spg_r1 = self.__pressure_grad_fpoint( ds_T.isel(r_dim=idx+1), ds_T_j1.isel(r_dim=idx+1),
                                        ds_T_i1.isel(r_dim=idx+1), ds_T_j1i1.isel(r_dim=idx+1), "u" )
                
                # average from f to u point and calculate velocities
                e2u_j1 = 0.5 * (ds_T_j1.isel(r_dim=idx).e2 + ds_T_j1i1.isel(r_dim=idx).e2 )
                u_hpg = -(0.5 * (self.data_F.e2[idx]*hpg/f[idx] + self.data_F.e2[idx+1]*hpg_r1/f[idx+1]) 
                                / (e2u_j1 * ref_density))
                u_spg = -(0.5 * (self.data_F.e2[idx]*spg/f[idx] + self.data_F.e2[idx+1]*spg_r1/f[idx+1]) 
                                / (e2u_j1 * ref_density))                
                normal_velocity_hpg[:,:,idx] = u_hpg.values
                normal_velocity_spg[:,idx] = u_spg.values
                horizontal_scale[:,idx] = e2u_j1
             
            # v flux (-v is positive across transect)
            elif dx[idx] > 0: 
                # calculate the pressure gradients at two f points defining a segment of the transect                                  
                hpg, spg = self.__pressure_grad_fpoint(ds_T.isel(r_dim=idx), ds_T_j1.isel(r_dim=idx),
                                        ds_T_i1.isel(r_dim=idx), ds_T_j1i1.isel(r_dim=idx), "v")
                hpg_r1, spg_r1 = self.__pressure_grad_fpoint(ds_T.isel(r_dim=idx+1), ds_T_j1.isel(r_dim=idx+1),
                                        ds_T_i1.isel(r_dim=idx+1),ds_T_j1i1.isel(r_dim=idx+1), "v")
                
                # average from f to v point and calculate velocities
                e1v_i1 = 0.5 * ( ds_T_i1.isel(r_dim=idx).e1 + ds_T_j1i1.isel(r_dim=idx).e1 )
                v_hpg = (0.5 * (self.data_F.e1[idx]*hpg/f[idx] + self.data_F.e1[idx+1]*hpg_r1/f[idx+1])
                                / (e1v_i1 * ref_density))
                v_spg = (0.5 * (self.data_F.e1[idx]*spg/f[idx] + self.data_F.e1[idx+1]*spg_r1/f[idx+1])
                                / (e1v_i1 * ref_density))
                normal_velocity_hpg[:,:,idx] = -v_hpg.values
                normal_velocity_spg[:,idx] = -v_spg.values
                horizontal_scale[:,idx] = e1v_i1
                
            # v flux (+v is positive across transect)    
            elif dx[idx] < 0:
                # calculate the pressure gradients at two f points defining a segment of the transect                                  
                hpg, spg = self.__pressure_grad_fpoint(ds_T.isel(r_dim=idx), ds_T_j1.isel(r_dim=idx),
                                        ds_T_i1.isel(r_dim=idx), ds_T_j1i1.isel(r_dim=idx), "v")
                hpg_r1, spg_r1 = self.__pressure_grad_fpoint(ds_T.isel(r_dim=idx+1), ds_T_j1.isel(r_dim=idx+1),
                                        ds_T_i1.isel(r_dim=idx+1),ds_T_j1i1.isel(r_dim=idx+1), "v")
                
                # average from f to v point and calculate velocities
                e1v = 0.5 * ( ds_T.isel(r_dim=idx).e1 + ds_T_j1.isel(r_dim=idx).e1 )
                v_hpg = (0.5 * (self.data_F.e1[idx]*hpg/f[idx] + self.data_F.e1[idx+1]*hpg_r1/f[idx+1])
                                / (e1v * ref_density))
                v_spg = (0.5 * (self.data_F.e1[idx]*spg/f[idx] + self.data_F.e1[idx+1]*spg_r1/f[idx+1])
                                / (e1v * ref_density))                   
                normal_velocity_hpg[:,:,idx] = v_hpg.values 
                normal_velocity_spg[:,idx] = v_spg.values 
                horizontal_scale[:,idx] = e1v
        
        normal_velocity_hpg[:,:,-1] = np.nan
        normal_velocity_spg[:,-1] = np.nan
        
        normal_velocity_hpg = np.where( ds_T.depth_z_levels.values[:,np.newaxis] <= ds_T.bathymetry.values, 
                               normal_velocity_hpg, np.nan )
        
        # remove redundent levels    
        active_z_levels = np.count_nonzero(~np.isnan(normal_velocity_hpg),axis=1).max() 
        normal_velocity_hpg = normal_velocity_hpg[:,:active_z_levels,:]
        z_levels = ds_T.depth_z_levels.values[:active_z_levels]
        
        coords_hpg={'depth_z_levels': (('depth_z_levels'), z_levels),
                'latitude': (('r_dim'), self.data_F.latitude),
                'longitude': (('r_dim'), self.data_F.longitude)}
        dims_hpg=['depth_z_levels', 'r_dim']
        attributes_hpg = {'units': 'm/s', 'standard name': 'velocity across the \
                          transect due to the hydrostatic pressure gradient'}
        coords_spg={'latitude': (('r_dim'), self.data_F.latitude),
                'longitude': (('r_dim'), self.data_F.longitude)}
        dims_spg=['r_dim']
        attributes_spg = {'units': 'm/s', 'standard name': 'velocity across the \
                          transect due to the surface pressure gradient'}
        
        if 't_dim' in ds_T.dims:
            coords_hpg['time'] = (('t_dim'), ds_T.time)
            dims_hpg.insert(0, 't_dim')
            coords_spg['time'] = (('t_dim'), ds_T.time)
            dims_spg.insert(0, 't_dim')
        
        self.data_tran['normal_velocity_hpg'] = xr.DataArray( np.squeeze(normal_velocity_hpg),
                coords=coords_hpg, dims=dims_hpg, attrs=attributes_hpg)
        self.data_tran['normal_velocity_spg'] = xr.DataArray( np.squeeze(normal_velocity_spg),
                coords=coords_spg, dims=dims_spg, attrs=attributes_spg)

        self.data_tran['transport_across_AB_hpg'] = ( self.data_tran
                .normal_velocity_hpg.fillna(0).integrate(dim='depth_z_levels') ) * horizontal_scale / 1000000   
        self.data_tran.transport_across_AB_hpg.attrs = {'units': 'Sv', 
                                  'standard_name': 'volume transport across transect due to the hydrostatic pressure gradient'}
        
        
        #depth_3d = self.data_tran.depth_z_levels.broadcast_like(self.data_tran.normal_velocity_hpg)
        #H = depth_3d.where(~self.data_tran.normal_velocity_hpg.to_masked_array().mask).max(dim='z_dim')
        H = ds_T.bathymetry.values
        self.data_tran['transport_across_AB_spg'] = self.data_tran.normal_velocity_spg * H * horizontal_scale / 1000000
        self.data_tran.transport_across_AB_spg.attrs = {'units': 'Sv', 
                                  'standard_name': 'volume transport across transect due to the surface pressure gradient'}
        
        return  # TODO Should this return something? If not the statement is not needed

    def construct_pressure( self, ds_T, ref_density, z_levels=None):   
        '''
            This method is for calculating the hydrostatic and surface pressure fields
            on z-levels. The motivation is to enable the calculation of horizontal 
            gradients and is primarily intended as an internal function; however, it can
            also be used with care to create pressure and density fields on a 
            transect defined on the t_grid.
            
            The ds_T argument is the t-grid dataset subsetted along a 
            transect. 
            Requirements: The supplied t-grid dataset must contain the 
            Practical Salinity and the Potential Temperature variables. The depth_0
            field must also be supplied. The GSW package is used to calculate
            The Absolute Pressure, Absolute Salinity and Conservate Temperature.
            
            Three new variables (density, hydrostatic pressure, surface pressure)
            are created and added to the ds_T dataset:
                density_zlevels       (t_dim, depth_z_levels, r_dim)
                pressure_h_zlevels    (t_dim, depth_z_levels, r_dim)
                pressure_s            (t_dim, r_dim)
            
            Note that density is constructed using the EOS10
            equation of state.

        Parameters
        ----------
        ds_T : xarray.Dataset
            the t-grid data defined on a transect.
        ref_density
            reference density value
        z_levels : (optional) numpy array
            1d array that defines the depths to interpolate the density and pressure
            on to.
        
        Returns
        -------
        None.

        '''
        # TODO Probably want a better description for this log message
        debug(f"calculating the hydrostatic and surface pressure fields on z-levels {z_levels} for {get_slug(self)}")
        if 't_dim' not in ds_T.dims:
            ds_T = ds_T.expand_dims(dim={'t_dim':1},axis=0)

        if z_levels is None:   
            bathymetry = xr.open_dataset( self.filename_domain ).bathy_metry.squeeze()
            z_levels_0_50 = np.arange(0,55,5.5)
            z_levels_60_200 = np.arange(60,210,10)
            z_levels_250_600 = np.arange(200,650,50)
            z_levels_650_ = np.arange(650,bathymetry.max()+150,150)
            z_levels = np.concatenate( (z_levels_0_50, z_levels_60_200, 
                                        z_levels_250_600, z_levels_650_) )
        
       # shape_ds = ( ds_T.t_dim.size, ds_T.z_dim.size, ds_T.r_dim.size )
        shape_ds = ( ds_T.t_dim.size, len(z_levels), ds_T.r_dim.size )
        salinity_z = np.ma.zeros( shape_ds )
        temperature_z = np.ma.zeros( shape_ds ) 
        salinity_s = ds_T.salinity.to_masked_array()
        temperature_s = ds_T.temperature.to_masked_array()
        s_levels = ds_T.depth_0.values
        
        # At the current time there does not appear to be a good algorithm for performing this 
        # type of interpolation without loops. Griddata is an option but does not
        # support extrapolation
        for it in ds_T.t_dim:
            for ir in ds_T.r_dim:
                if not np.all(np.isnan(salinity_s[it,:,ir].data)):  
                    # Need to remove the levels below the (envelope) bathymetry which are NaN
                    salinity_s_r = salinity_s[it,:,ir].compressed()
                    temperature_s_r = temperature_s[it,:,ir].compressed()
                    s_levels_r = s_levels[:len(salinity_s_r),ir]
                    
                    sal_func = interpolate.interp1d( s_levels_r, salinity_s_r, 
                                 kind='linear', fill_value="extrapolate")
                    temp_func = interpolate.interp1d( s_levels_r, temperature_s_r, 
                                 kind='linear', fill_value="extrapolate")
                    
                    salinity_z[it,:,ir] = sal_func(z_levels)
                    temperature_z[it,:,ir] = temp_func(z_levels)
                    # set levels below the bathymetry to nan
                   # salinity_z[it,:,ir] = np.where( z_levels <= ds_T.bathymetry.values[ir], 
                   #             sal_func(z_levels), np.nan )
                   # temperature_z[it,:,ir] = np.where( z_levels <= ds_T.bathymetry.values[ir], 
                   #             temp_func(z_levels), np.nan ) 
                    

        # Absolute Pressure    
        pressure_absolute = np.ma.masked_invalid(
            gsw.p_from_z( -z_levels[:,np.newaxis], ds_T.latitude ) ) # depth must be negative           
        # Absolute Salinity           
        salinity_absolute = np.ma.masked_invalid(
            gsw.SA_from_SP( salinity_z, pressure_absolute, ds_T.longitude, ds_T.latitude ) )
        salinity_absolute = np.ma.masked_less(salinity_absolute,0)
        # Conservative Temperature
        temp_conservative = np.ma.masked_invalid(
            gsw.CT_from_pt( salinity_absolute, temperature_z ) )
        # In-situ density
        density_z = np.ma.masked_invalid( gsw.rho( 
            salinity_absolute, temp_conservative, pressure_absolute ) )
        
                        
        coords={'depth_z_levels': (('depth_z_levels'), z_levels),
                'latitude': (('r_dim'), ds_T.latitude),
                'longitude': (('r_dim'), ds_T.longitude)}
        dims=['depth_z_levels', 'r_dim']
        attributes = {'units': 'kg / m^3', 'standard name': 'In-situ density on the z-level vertical grid'}
        
        if shape_ds[0] != 1:
            coords['time'] = (('t_dim'), ds_T.time.values)
            dims.insert(0, 't_dim')
          
        ds_T['density_zlevels'] = xr.DataArray( np.squeeze(density_z), 
                coords=coords, dims=dims, attrs=attributes )
    
        # cumulative integral of density on z levels
        # Note that zero density flux is assumed at z=0
        density_cumulative = -cumtrapz( density_z, x=-z_levels, axis=1, initial=0)

        hydrostatic_pressure = density_cumulative * self.GRAVITY
        
        attributes = {'units': 'kg m^{-1} s^{-2}', 'standard name': 'Hydrostatic pressure on the z-level vertical grid'}
        ds_T['pressure_h_zlevels'] = xr.DataArray( np.squeeze(hydrostatic_pressure), 
                coords=coords, dims=dims, attrs=attributes )
        
        ds_T['pressure_s'] = ref_density * self.GRAVITY * ds_T.ssh.squeeze()
        ds_T.pressure_s.attrs = {'units': 'kg m^{-1} s^{-2}', 
                                  'standard_name': 'surface pressure'}
        
        return

    def moving_average(self, array_to_smooth, window=2, axis=-1):  # TODO This could be a static method
        '''
        Returns the input array smoothed along the given axis using convolusion
        '''
        debug(f"Fetching moving average for {array_to_smooth}")
        return convolve1d( array_to_smooth, np.ones(window), axis=axis ) / window
    
    def interpolate_slice(self, variable_slice, depth, interpolated_depth=None ):  # TODO This could be a static method
        '''
        Linearly interpolates the variable at a single time along the z_dim, which must be the
        first axis.

        Parameters
        ----------
        variable_slice : Variable to interpolate (z_dim, transect_dim)
        depth : The depth at each z point for each point along the transect
        interpolated_depth : (optional) desired depth profile to interpolate to. If not supplied
            a uniform depth profile uniformaly spaced between zero and variable max depth will be used
            with a spacing of 2 metres.

        Returns
        -------
        interpolated_depth_variable_slice : Interpolated variable
        interpolated_depth : Interpolation depth

        '''
        debug(f"Interpolating slice {variable_slice} at depths {depth}")
        if interpolated_depth is None:
            interpolated_depth = np.arange(0, np.nanmax(depth), 2)
            
        interpolated_depth_variable_slice = np.zeros( (len(interpolated_depth), variable_slice.shape[-1]) )
        for i in np.arange(0, variable_slice.shape[-1] ):
            depth_func = interpolate.interp1d( depth[:,i], variable_slice[:,i], axis=0, bounds_error=False )        
            interpolated_depth_variable_slice[:,i] = depth_func( interpolated_depth )
            
        return (interpolated_depth_variable_slice, interpolated_depth )  # TODO Brackets aren't required
    
    def plot_transect_on_map(self):
        '''
        Plot transect location on a map
        
        Example usage:
        --------------
        tran = coast.Transect( (54,-15), (56,-12), nemo_f, nemo_t, nemo_u, nemo_v )
        tran.plot_map()
        '''
        debug(f"Generating plot on map for {get_slug(self)}")
        try:
            import cartopy.crs as ccrs  # mapping plots
            import cartopy.feature  # add rivers, regional boundaries etc
            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
            from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
        except ImportError:
            import sys
            warnings.warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
            sys.exit(-1)

        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca()
        ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())

        cset = plt.plot(self.data_F.longitude, self.data_F.latitude, c='k')

        ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
        coast = NaturalEarthFeature(category='physical', scale='50m',
                                    facecolor=[0.8,0.8,0.8], name='coastline',
                                    alpha=0.5)
        ax.add_feature(coast, edgecolor='gray')

        gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                          linewidth=0.5, color='gray', alpha=0.5, linestyle='-')

        gl.top_labels = False
        gl.bottom_labels = True
        gl.right_labels = False
        gl.left_labels = True
        gl.xformatter = LONGITUDE_FORMATTER
        gl.yformatter = LATITUDE_FORMATTER
        
        plt.title('Map of transect location')
        #plt.show() # Can only adjust axis if fig is plotted already..
        
        return fig, ax
    
    def plot_normal_velocity(self, time, plot_info: dict, cmap, smoothing_window=0):  
        '''
            Quick plot routine of velocity across the transect AB at a specific time.
            An option is provided to smooth the velocities along the transect.
            NOTE: For smoothing use even integers to smooth the x and y velocities together
    
    
    
        Parameters
        ---------------
        time: either as integer index or actual time as a string.
        plot_info: dictionary of infomation {'fig_size': value, 'title': value, 'vmin':value, 'vmax':value}
        Note that if vmin and max are not set then the colourbar will be centred at zero
        smoothing_window: smoothing via convolusion, larger number applies greater smoothing, recommended
        to use even integers

        
        '''
        debug(f"Plotting normal velocity for {get_slug(self)} with plot_info {plot_info}")
        try:
            data = self.data_tran.sel(t_dim = time)
        except KeyError:
            data = self.data_tran.isel(t_dim = time)        
        
        if smoothing_window != 0:
            normal_velocities, depth = self.interpolate_slice( data.normal_velocities, data.depth_0 )            
            normal_velocities = self.moving_average(normal_velocities, smoothing_window, axis=-1)
            r_dim_2d = np.broadcast_to( data.r_dim, normal_velocities.shape  )
        else:
            normal_velocities = data.normal_velocities
            depth = data.depth_0
            _ , r_dim_2d = xr.broadcast( depth, data.r_dim  )
                    
        import matplotlib.pyplot as plt
        plt.close('all')
        fig = plt.figure(figsize=plot_info['fig_size'])
        ax = fig.gca()

        plt.pcolormesh(r_dim_2d, depth, normal_velocities, cmap=cmap)
            
        plt.title(plot_info['title'])
        plt.ylabel('Depth [m]')
        try:
            plt.clim(vmin=plot_info['vmin'], vmax=plot_info['vmax'])
        except KeyError:
            lim = np.nanmax(np.abs(normal_velocities))
            plt.clim(vmin=-lim, vmax=lim)
        plt.xticks([0,data.r_dim.values[-1]],['A','B'])
        plt.colorbar(label='Velocities across AB [m/s]')
        plt.gca().invert_yaxis()

        plt.show()
        return fig,ax

    def plot_depth_integrated_transport(self, time, plot_info: dict, smoothing_window=0):
        '''
            Quick plot routine of depth integrated transport across the transect AB at a specific time.
            An option is provided to smooth along the transect via convolution, 
            NOTE: For smoothing use even integers to smooth the x and y velocities together
    
        Parameters
        ---------------
        time: either as integer index or actual time as a string.
        plot_info: dictionary of infomation {'fig_size': value, 'title': value}
        smoothing_window: smoothing via convolusion, larger number applies greater smoothing. 
        Recommended to use even integers.
        returns: pyplot object
        '''
        debug(f"Generating quick plot for {get_slug(self)} with plot_info {plot_info}")
        try:
            data = self.data_tran.sel(t_dim = time)
        except KeyError:
            data = self.data_tran.isel(t_dim = time)            
        
        if smoothing_window != 0:    
            transport = self.moving_average(data.depth_integrated_transport_across_AB, smoothing_window, axis=-1)
        else:
            transport = data.depth_integrated_transport_across_AB
        
        import matplotlib.pyplot as plt
        plt.close('all')
        fig = plt.figure(figsize=plot_info['fig_size'])
        ax = fig.gca()

        plt.plot( data.r_dim, transport )

        plt.title(plot_info['title'])
        plt.xticks([0,data.r_dim[-1]],['A','B'])
        plt.ylabel('Volume transport across AB [SV]')
        plt.show()
        return fig,ax

    def construct_density_on_z_levels( self, EOS='EOS10'):#, z_levels=None ):        
        '''
            For s-level model output this method recontructs the in-situ density 
            onto z_levels along the transect. The z_levels and density field
            are added to the data_T dataset attribute. 
            
            Requirements: The supplied t-grid dataset must contain the 
            Practical Salinity and the Potential Temperature variables. The depth_0
            field must also be supplied. The GSW package is used to calculate
            The Absolute Pressure, Absolute Salinity and Conservate Temperature.
            
            This method is useful when horizontal gradients of the density field 
            are required. Currently z_levels cannot be specified, the 
            method will contruct a z_levels profile from the s_levels.
            
            Note that currently density can only be constructed using the EOS10
            equation of state.

        Parameters
        ----------
        EOS : equation of state, optional
            DESCRIPTION. The default is 'EOS10'.


        Returns
        -------
        None.
        adds attributes Transect.data_T.depth_z_levels and Transect.data_T.density_z_levels

        '''
        debug(f"Constructing in-situ density on z-levels for {get_slug(self)} with EOS \"{EOS}\"")
        try:
            if EOS != 'EOS10': 
                raise ValueError(get_slug(self) + ': Density calculation for ' + EOS + ' not implemented.')
            if self.data_T is None:
                raise ValueError(get_slug(self) + ': Density calculation can only be performed \
                    when a t-grid object has been assigned to the nemo_T attribute. This\
                    can be done at initialisation.' )
         # TODO Should this be reinstated and converted to a log message?
          #  if not self.data_T.ln_sco.item():
           #     raise ValueError(str(self) + ': Density calculation only implemented for s-vertical-coordinates.')            
    
            #if z_levels is None:
            z_levels = self.data_T.depth_0.max(dim=(['r_dim']))                
            z_levels_min = self.data_T.depth_0[0,:].max(dim=(['r_dim']))
            z_levels[0] = z_levels_min
            #else:
             #   z_max = self.data_T.depth_0.max(dim=(['r_dim','z_dim'])).item()
              #  z_min = self.data_T.depth_0[0,:].max(dim=(['r_dim'])).item()
               # z_levels = z_levels[z_levels<=z_max]
                #z_levels = z_levels[z_levels>=z_min]                     
            
            try:    
                shape_ds = ( self.data_T.t_dim.size, z_levels.size, 
                                self.data_T.r_dim.size )
                sal = self.data_T.salinity.to_masked_array()
                temp = self.data_T.temperature.to_masked_array()                
            except AttributeError:
                shape_ds = ( 1, z_levels.size, self.data_T.r_dim.size )
                sal = self.data_T.salinity.to_masked_array()[np.newaxis,...]
                temp = self.data_T.temperature.to_masked_array()[np.newaxis,...]
            
            sal_z_levels = np.ma.zeros( shape_ds )
            temp_z_levels = np.ma.zeros( shape_ds )
            density_z_levels = np.ma.zeros( shape_ds )
            
            s_levels = self.data_T.depth_0.to_masked_array()
            lat = self.data_T.latitude.values
            lon = self.data_T.longitude.values
    
            for it in np.arange(0, shape_ds[0]):
                for ir in self.data_T.r_dim:
                    if np.all(np.isnan(sal[it,:,ir])):
                        density_z_levels[it,:,ir] = np.nan
                        density_z_levels[it,:,ir].mask = True
                    else:                      
                        sal_func = interpolate.interp1d( s_levels[:,ir], sal[it,:,ir], 
                                    bounds_error=False, kind='linear')
                        temp_func = interpolate.interp1d( s_levels[:,ir], temp[it,:,ir], 
                                    bounds_error=False, kind='linear')
                        
                        sal_z_levels[it,:,ir] = sal_func(z_levels.values)
                        temp_z_levels[it,:,ir] = temp_func(z_levels.values)
                        
            # Absolute Pressure    
            pressure_absolute = np.ma.masked_invalid(
                gsw.p_from_z( -z_levels.values[:,np.newaxis], lat ) ) # depth must be negative           
            # Absolute Salinity           
            sal_absolute = np.ma.masked_invalid(
                gsw.SA_from_SP( sal_z_levels, pressure_absolute, lon, lat ) )
            sal_absolute = np.ma.masked_less(sal_absolute,0)
            # Conservative Temperature
            temp_conservative = np.ma.masked_invalid(
                gsw.CT_from_pt( sal_absolute, temp_z_levels ) )
            # In-situ density
            density_z_levels = np.ma.masked_invalid( gsw.rho( 
                sal_absolute, temp_conservative, pressure_absolute ) )
            
            coords={'depth_z_levels': (('z_dim'), z_levels.values),
                    'latitude': (('r_dim'), self.data_T.latitude.values),
                    'longitude': (('r_dim'), self.data_T.longitude.values)}
            dims=['z_dim', 'r_dim']
            attributes = {'units': 'kg / m^3', 'standard name': 'In-situ density on the z-level vertical grid'}
            
            if shape_ds[0] != 1:
                coords['time'] = (('t_dim'), self.data_T.time.values)
                dims.insert(0, 't_dim')
              
            self.data_T['density_z_levels'] = xr.DataArray( np.squeeze(density_z_levels), 
                    coords=coords, dims=dims, attrs=attributes )

        except AttributeError as err:
            error(err)
        return



from dask import delayed
from dask import array
import xarray as xr
import numpy as np
from dask.distributed import Client
from warnings import warn
import copy
import scipy as sp
from .logging_util import get_slug, debug, info, warn, error
import sklearn.neighbors as nb

def calculate_haversine_distance(lon1, lat1, lon2, lat2):
    '''
    # Estimation of geographical distance using the Haversine function.
    # Input can be single values or 1D arrays of locations. This
    # does NOT create a distance matrix but outputs another 1D array.
    # This works for either location vectors of equal length OR a single loc
    # and an arbitrary length location vector.
    #
    # lon1, lat1 :: Location(s) 1.
    # lon2, lat2 :: Location(s) 2.
    '''

    # Convert to radians for calculations
    lon1 = xr.ufuncs.deg2rad(lon1)
    lat1 = xr.ufuncs.deg2rad(lat1)
    lon2 = xr.ufuncs.deg2rad(lon2)
    lat2 = xr.ufuncs.deg2rad(lat2)

    # Latitude and longitude differences
    dlat = (lat2 - lat1) / 2
    dlon = (lon2 - lon1) / 2

    # Haversine function.
    distance = xr.ufuncs.sin(dlat) ** 2 + xr.ufuncs.cos(lat1) * \
               xr.ufuncs.cos(lat2) * xr.ufuncs.sin(dlon) ** 2
    distance = 2 * 6371.007176 * xr.ufuncs.arcsin(xr.ufuncs.sqrt(distance))

    return distance

def remove_indices_by_mask(A, mask):
    '''
    Removes indices from a 2-dimensional array, A, based on true elements of
    mask. A and mask variable should have the same shape.
    '''
    A = np.array(A).flatten()
    mask = np.array(mask, dtype=bool).flatten()
    array_removed = A[~mask]
        
    return array_removed

def reinstate_indices_by_mask(array_removed, mask, fill_value=np.nan):
    '''
    Rebuilds a 2D array from a 1D array created using remove_indices_by_mask().
    False elements of mask will be populated using array_removed. MAsked
    indices will be replaced with fill_value
    '''
    array_removed = np.array(array_removed)
    original_shape = mask.shape
    mask = np.array(mask, dtype=bool).flatten()
    A = np.zeros(mask.shape)
    A[~mask] = array_removed
    A[mask] = fill_value
    A = A.reshape(original_shape)
    return A

def nearest_xy_indices(mod_lon, mod_lat, new_lon, new_lat, 
                       mask = None):
    '''
    Obtains the x and y indices of the nearest model points to specified
    lists of longitudes and latitudes. Makes use of sklearn.neighbours
    and its BallTree haversine method. 
    
    Example Useage
    ----------
    # Get indices of model points closest to altimetry points
    ind_x, ind_y = nemo.nearest_indices(altimetry.dataset.longitude,
                                        altimetry.dataset.latitude)
    # Nearest neighbour interpolation of model dataset to these points
    interpolated = nemo.dataset.isel(x_dim = ind_x, y_dim = ind_y)
    
    Parameters
    ----------
    mod_lon (2D array): Model longitude (degrees) array (2-dimensional)
    mod_lat (2D array): Model latitude (degrees) array (2-dimensions)
    new_lon (1D array): Array of longitudes (degrees) to compare with model
    new_lat (1D array): Array of latitudes (degrees) to compare with model
    mask (2D array): Mask array. Where True (or 1), elements of mod_lons
                     and mod lats will be removed prior to finding nearest
                     neighbours. e.g. if the nearest ocean point is 
        
    Returns
    -------
    Array of x indices, Array of y indices
    '''
    # Cast lat/lon to numpy arrays in case xarray things
    new_lon = np.array(new_lon)
    new_lat = np.array(new_lat)
    mod_lon = np.array(mod_lon)
    mod_lat = np.array(mod_lat)
    original_shape = mod_lon.shape
    
    # If a mask is supplied, remove indices from arrays.
    if mask is None:
        mod_lon = mod_lon.flatten()
        mod_lat = mod_lat.flatten()
    else:
        mod_lon[mask] = np.nan
        mod_lat[mask] = np.nan
        mod_lon = mod_lon.flatten()
        mod_lat = mod_lat.flatten()
    
    # Put lons and lats into 2D location arrays for BallTree: [lat, lon]
    mod_loc = np.vstack((mod_lat, mod_lon)).transpose()
    new_loc = np.vstack((new_lat, new_lon)).transpose()
    
    # Convert lat/lon to radians for BallTree
    mod_loc = np.radians(mod_loc)
    new_loc = np.radians(new_loc)
    
    # Do nearest neighbour interpolation using BallTree (gets indices)
    tree = nb.BallTree(mod_loc, leaf_size=5, metric='haversine')
    _, ind_1d = tree.query(new_loc, k=1)
    
    # Get 2D indices from 1D index output from BallTree
    ind_y, ind_x = np.unravel_index(ind_1d, original_shape)
    ind_x = xr.DataArray(ind_x.squeeze())
    ind_y = xr.DataArray(ind_y.squeeze())
    return ind_x, ind_y

def dataarray_time_slice(data_array, date0, date1):
    ''' Takes an xr.DataArray object and returns a new object with times 
    sliced between dates date0 and date1. date0 and date1 may be a string or
    datetime type object.'''
    if date0 is None and date1 is None:
        return data_array
    else: 
        data_array_sliced = data_array.swap_dims({'t_dim':'time'})
        time_max = data_array.time.max().values
        time_min = data_array.time.min().values
        if date0 is None:
            date0 = time_min
        if date1 is None:
            date1 = time_max
        data_array_sliced = data_array_sliced.sel(time = slice(date0, date1))
        data_array_sliced = data_array_sliced.swap_dims({'time':'t_dim'})
        return data_array_sliced


'''
Python definitions used to help with plotting routines.

*Methods Overview*
    -> geo_scatter(): Geographical scatter plot.
'''

import matplotlib.pyplot as plt
from warnings import warn
from .logging_util import get_slug, debug, info, warn, error
            
def geo_scatter(longitude, latitude, colors=None, 
                title='', xlim=None, ylim=None):
    '''
    Uses CartoPy to create a geographical scatter plot with land boundaries.
    
        Parameters
        ----------
        longitude : (array) Array of longitudes of marker locations
        latitude  : (array) Array of latitudes of marker locations
        colors    : (array) Array of values to use for colouring markers
        title     : (str) Plot title, to appear at top of figure
        xlim      : (tuple) Tuple of limits to apply to the x-axis (longitude axis)
        ylim      : (tuple) Limits to apply to the y-axis (latitude axis)
    
        Returns
        -------
        Figure and axis objects for further customisation
    
    '''
    try:
        import cartopy.crs as ccrs  # mapping plots
        import cartopy.feature  # add rivers, regional boundaries etc
        from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
        from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
    except ImportError:
        import sys
        warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
        sys.exit(-1)

    fig = plt.figure(figsize=(10, 10))
    ax = fig.gca()
    ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())

    if colors is None:
        plt.scatter(longitude, latitude)
    else:
        plt.scatter(longitude, y=latitude, c = colors)

    ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
    coast = NaturalEarthFeature(category='physical', scale='50m',
                                facecolor=[0.8,0.8,0.8], name='coastline',
                                alpha=0.5)
    ax.add_feature(coast, edgecolor='gray')

    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                      linewidth=0.5, color='gray', alpha=0.5, linestyle='-')
    gl.top_labels = False
    gl.bottom_labels = True
    gl.right_labels = False
    gl.left_labels = True
    gl.xformatter = LONGITUDE_FORMATTER
    gl.yformatter = LATITUDE_FORMATTER
    plt.title(title)
    plt.colorbar()
    
    if xlim is not None:
        plt.xlim(xlim)
    if ylim is not None:
        plt.ylim(ylim)
    
    plt.show()
    return fig, ax


from .COAsT import COAsT
import numpy as np
from .logging_util import get_slug, debug, info, warn, warning, error


class OBSERVATION(COAsT):
    def set_dimension_mapping(self):
        self.dim_mapping = None
        debug(f"{get_slug(self)} dim_mapping set to {self.dim_mapping}")

    def adjust_longitudes(self, lonbounds=[-180,180]):  # TODO This [list] should probably be a (tuple)
        debug(f"Adjusting {get_slug(self)} longitudes with lonbounds {lonbounds}")
        bool0 = self['longitude']<lonbounds[0]
        bool1 = self['longitude']>lonbounds[1]
        self['longitude'][bool0] = self['longitude'][bool0] + 360
        self['longitude'][bool1] = self['longitude'][bool1] - 360



from .COAsT import COAsT  # ???
import numpy as np
import xarray as xr
from .logging_util import get_slug, debug, error, info
import sklearn.metrics as metrics
from . import general_utils, plot_util, crps_util

class ALTIMETRY(COAsT):
    '''
    An object for reading, storing and manipulating altimetry data.
    Currently the object contains functionality for reading altimetry netCDF 
    data from the CMEMS database. This is the default for initialisation.
    
    Data should be stored in an xarray.Dataset, in the form:
        
    * Date Format Overview *
        
        1. A single dimension (time).
        2. Three coordinates: time, latitude, longitude. All lie on the 't_dim'
           dimension.
        3. Observed variable DataArrays on the t_dim dimension.
        
    There are currently no naming conventions for the variables however
    examples from the CMEMS database include sla_filtered, sla_unfiltered and
    mdt (mean dynamic topography).
    
    * Methods Overview *
        
        *Initialisation and File Reading*
        -> __init__(): Initialises an ALTIMETRY object.
        -> read_cmems(): Reads data from a CMEMS netCDF file.
        
        *Plotting*
        -> quick_plot(): Makes a quick along-track plot of specified data.
        
        *Model Comparison*
        -> obs_operator(): For interpolating model data to this object.
        -> cprs(): Calculates the CRPS between a model and obs variable.
        -> difference(): Differences two specified variables
        -> absolute_error(): Absolute difference, two variables
        -> mean_absolute_error(): MAE between two variables
        -> root_mean_square_error(): RMSE between two variables
        -> time_mean(): Mean of a variable in time
        -> time_std(): St. Dev of a variable in time
        -> time_correlation(): Correlation between two variables
        -> time_covariance(): Covariance between two variables
        -> basic_stats(): Calculates multiple of the above metrics.

    '''
##############################################################################
###                ~ Initialisation and File Reading ~                     ###
##############################################################################

    def __init__(self, file=None, chunks: dict=None, multiple=False):
        debug(f"Creating a new {get_slug(self)}")
        if file is not None:
            self.read_cmems(file, chunks, multiple)
        else:
            self.dataset = None
        debug(f"{get_slug(self)} initialised")
        return
    
    def read_cmems(self, file, chunks, multiple):
        ''' Reads altimetry data from a CMEMS netcdf file. Calls COAsT.init()
        to make use of its load methods'''
        super().__init__(file, chunks, multiple)
        self.dataset = self.dataset.rename_dims(self.dim_mapping)
        #self.dataset.attrs = {}

    def set_dimension_mapping(self):
        self.dim_mapping = {'time': 't_dim'}
        debug(f"{get_slug(self)} dim_mapping set to {self.dim_mapping}")

    def set_variable_mapping(self):
        self.var_mapping = None
        debug(f"{get_slug(self)} var_mapping set to {self.var_mapping}")
        
    def subset_indices_lonlat_box(self, lonbounds, latbounds):
        """Generates array indices for data which lies in a given lon/lat box.

        Keyword arguments:
        lon       -- Longitudes, 1D or 2D.
        lat       -- Latitudes, 1D or 2D
        lonbounds -- Array of form [min_longitude=-180, max_longitude=180]
        latbounds -- Array of form [min_latitude, max_latitude]
        
        return: Indices corresponding to datapoints inside specified box
        """
        debug(f"Subsetting {get_slug(self)} indices in {lonbounds}, {latbounds}")
        lon = self.dataset.longitude.copy()
        lat = self.dataset.latitude
        lon[lon>180] = lon[lon>180] - 360
        lon[lon<-180] = lon[lon<-180] + 360
        ff1 = ( lon > lonbounds[0] ).astype(int)  # FIXME This should fail? We can just treat bools as ints here...
        ff2 = ( lon < lonbounds[1] ).astype(int)
        ff3 = ( lat > latbounds[0] ).astype(int)
        ff4 = ( lat < latbounds[1] ).astype(int)
        indices = np.where( ff1 * ff2 * ff3 * ff4 )
        return indices[0]

##############################################################################
###                ~            Plotting             ~                     ###
##############################################################################

    def quick_plot(self, color_var_str: str=None):
        '''
        '''
        
        if color_var_str is not None:
            color_var = self.dataset[color_var_str]
            title = color_var_str
        else:
            color_var = None
            title = 'Altimetry observation locations'
        info("Drawing a quick plot...")
        fig, ax =  plot_util.geo_scatter(self.dataset.longitude, 
                                         self.dataset.latitude,
                                         color_var, title=title )
        info("Plot ready, displaying!")
        return fig, ax

##############################################################################
###                ~        Model Comparison         ~                     ###
##############################################################################
    
    def obs_operator(self, model, mod_var_name:str, 
                                time_interp = 'nearest'):
        '''
        For interpolating a model dataarray onto altimetry locations and times.
        
        For ALTIMETRY, the interpolation is done independently in two steps:
            1. Horizontal space
            2. Time
        Model data is taken at the surface if necessary (0 index). 
    
        Example usage:
        --------------
        altimetry.obs_operator(nemo_obj, 'sossheig')

        Parameters
        ----------
        model : model object (e.g. NEMO)
        mod_var: variable name string to use from model object
        time_interp: time interpolation method (optional, default: 'nearest')
            This can take any string scipy.interpolate would take. e.g.
            'nearest', 'linear' or 'cubic'
        Returns
        -------
        Adds a DataArray to self.dataset, containing interpolated values.
        '''

        debug(f"Interpolating {get_slug(model)} \"{mod_var_name}\" with time_interp \"{time_interp}\"")

        # Get data arrays
        mod_var = model.dataset[mod_var_name]
        
        # Depth interpolation -> for now just take 0 index
        if 'z_dim' in mod_var.dims:
            mod_var = mod_var.isel(z_dim=0).squeeze()
        
        # Cast lat/lon to numpy arrays
        obs_lon = np.array(self.dataset.longitude).flatten()
        obs_lat = np.array(self.dataset.latitude).flatten()
        
        interpolated = model.interpolate_in_space(mod_var, obs_lon, 
                                                        obs_lat)
        
        # Interpolate in time if t_dim exists in model array
        if 't_dim' in mod_var.dims:
            interpolated = model.interpolate_in_time(interpolated, 
                                                     self.dataset.time,
                                                     interp_method=time_interp)
            # Take diagonal from interpolated array (which contains too many points)
            diag_len = interpolated.shape[0]
            diag_ind = xr.DataArray(np.arange(0, diag_len))
            interpolated = interpolated.isel(interp_dim=diag_ind, t_dim=diag_ind)
            interpolated = interpolated.swap_dims({'dim_0':'t_dim'})

        # Store interpolated array in dataset
        new_var_name = 'interp_' + mod_var_name
        self.dataset[new_var_name] = interpolated
    
    def crps(self, model_object, model_var_name, obs_var_name, 
             nh_radius: float = 20, cdf_type:str='empirical', 
             time_interp:str='linear', create_new_object = True):
        
        '''
        Comparison of observed variable to modelled using the Continuous
        Ranked Probability Score. This is done using this ALTIMETRY object.
        This method specifically performs a single-observation neighbourhood-
        forecast method.
        
        Parameters
        ----------
        model_object (model) : Model object (NEMO) containing model data
        model_var_name (str) : Name of model variable to compare.
        obs_var_name (str)   : Name of observed variable to compare.
        nh_radius (float)    : Neighbourhood rad
        cdf_type (str)       : Type of cumulative distribution to use for the
                               model data ('empirical' or 'theoretical').
                               Observations always use empirical.
        time_interp (str)    : Type of time interpolation to use (s)
        create_new_obj (bool): If True, save output to new ALTIMETRY obj.
                               Otherwise, save to this obj.
          
        Returns
        -------
        xarray.Dataset containing times, sealevel and quality control flags
        
        Example Useage
        -------
        # Compare modelled 'sossheig' with 'sla_filtered' using CRPS
        crps = altimetry.crps(nemo, 'sossheig', 'sla_filtered')
        '''
        
        mod_var = model_object.dataset[model_var_name]
        obs_var = self.dataset[obs_var_name]
        
        crps_list, n_model_pts, contains_land = crps_util.crps_sonf_moving( 
                               mod_var, 
                               obs_var.longitude.values, 
                               obs_var.latitude.values, 
                               obs_var.values, 
                               obs_var.time.values, 
                               nh_radius, cdf_type, time_interp )
        if create_new_object:
            new_object = ALTIMETRY()
            new_dataset = self.dataset[['longitude','latitude','time']]
            new_dataset['crps'] =  (('t_dim'),crps_list)
            new_dataset['crps_n_model_pts'] = (('t_dim'), n_model_pts)
            new_dataset['crps_contains_land'] = (('t_dim'), contains_land)
            new_object.dataset = new_dataset
            return new_object
        else:
            self.dataset['crps'] =  (('t_dim'),crps_list)
            self.dataset['crps_n_model_pts'] = (('t_dim'), n_model_pts)
            self.dataset['crps_contains_land'] = (('t_dim'), contains_land)
    
    def difference(self, var_str0:str, var_str1:str, date0=None, date1=None):
        ''' Difference two variables defined by var_str0 and var_str1 between
        two dates date0 and date1. Returns xr.DataArray '''
        var0 = self.dataset[var_str0]
        var1 = self.dataset[var_str1]
        var0 = general_utils.dataarray_time_slice(var0, date0, date1).values
        var1 = general_utils.dataarray_time_slice(var1, date0, date1).values
        diff = var0 - var1
        return xr.DataArray(diff, dims='t_dim', name='error',
                            coords={'time':self.dataset.time})
    
    def absolute_error(self, var_str0, var_str1, date0=None, date1=None):
        ''' Absolute difference two variables defined by var_str0 and var_str1 
        between two dates date0 and date1. Return xr.DataArray '''
        var0 = self.dataset[var_str0]
        var1 = self.dataset[var_str1]
        var0 = general_utils.dataarray_time_slice(var0, date0, date1).values
        var1 = general_utils.dataarray_time_slice(var1, date0, date1).values
        adiff = np.abs(var0 - var1)
        return xr.DataArray(adiff, dims='t_dim', name='absolute_error', 
                            coords={'time':self.dataset.time})
    
    def mean_absolute_error(self, var_str0, var_str1, date0=None, date1=None):
        ''' Mean absolute difference two variables defined by var_str0 and 
        var_str1 between two dates date0 and date1. Return xr.DataArray '''
        var0 = self.dataset[var_str0]
        var1 = self.dataset[var_str1]
        var0 = general_utils.dataarray_time_slice(var0, date0, date1).values
        var1 = general_utils.dataarray_time_slice(var1, date0, date1).values
        mae = metrics.mean_absolute_error(var0, var1)
        return mae
    
    def root_mean_square_error(self, var_str0, var_str1, date0=None, date1=None):
        ''' Root mean square difference two variables defined by var_str0 and 
        var_str1 between two dates date0 and date1. Return xr.DataArray '''
        var0 = self.dataset[var_str0]
        var1 = self.dataset[var_str1]
        var0 = general_utils.dataarray_time_slice(var0, date0, date1).values
        var1 = general_utils.dataarray_time_slice(var1, date0, date1).values
        rmse = metrics.mean_squared_error(var0, var1)
        return np.sqrt(rmse)
    
    def time_mean(self, var_str, date0=None, date1=None):
        ''' Time mean of variable var_str between dates date0, date1'''
        var = self.dataset[var_str]
        var = general_utils.dataarray_time_slice(var, date0, date1)
        return np.nanmean(var)
    
    def time_std(self, var_str, date0=None, date1=None):
        ''' Time st. dev of variable var_str between dates date0 and date1'''
        var = self.dataset[var_str]
        var = general_utils.dataarray_time_slice(var, date0, date1)
        return np.nanstd(var)
    
    def time_correlation(self, var_str0, var_str1, date0=None, date1=None, 
                         method='pearson'):
        ''' Time correlation between two variables defined by var_str0, 
        var_str1 between dates date0 and date1. Uses Pandas corr().'''
        var0 = self.dataset[var_str0]
        var1 = self.dataset[var_str1]
        var0 = var0.rename('var1')
        var1 = var1.rename('var2')
        var0 = general_utils.dataarray_time_slice(var0, date0, date1)
        var1 = general_utils.dataarray_time_slice(var1, date0, date1)
        pdvar = xr.merge((var0, var1))
        pdvar = pdvar.to_dataframe()
        corr = pdvar.corr(method=method)
        return corr.iloc[0,1]
    
    def time_covariance(self, var_str0, var_str1, date0=None, date1=None):
        ''' Time covariance between two variables defined by var_str0, 
        var_str1 between dates date0 and date1. Uses Pandas corr().'''
        var0 = self.dataset[var_str0]
        var1 = self.dataset[var_str1]
        var0 = var0.rename('var1')
        var1 = var1.rename('var2')
        var0 = general_utils.dataarray_time_slice(var0, date0, date1)
        var1 = general_utils.dataarray_time_slice(var1, date0, date1)
        pdvar = xr.merge((var0, var1))
        pdvar = pdvar.to_dataframe()
        cov = pdvar.cov()
        return cov.iloc[0,1]
    
    def basic_stats(self, var_str0, var_str1, date0 = None, date1 = None,
                    create_new_object = True):
        ''' Calculates a selection of statistics for two variables defined by
        var_str0 and var_str1, between dates date0 and date1. This will return
        their difference, absolute difference, mean absolute error, root mean 
        square error, correlation and covariance. If create_new_object is True
        then this method returns a new ALTIMETRY object containing statistics,
        otherwise variables are saved to the dateset inside this object. '''
        
        diff = self.difference(var_str0, var_str1, date0, date1)
        ae = self.absolute_error(var_str0, var_str1, date0, date1)
        mae = self.mean_absolute_error(var_str0, var_str1, date0, date1)
        rmse = self.root_mean_square_error(var_str0, var_str1, date0, date1)
        corr = self.time_correlation(var_str0, var_str1, date0, date1)
        cov = self.time_covariance(var_str0, var_str1, date0, date1)
        
        if create_new_object:
            new_object = ALTIMETRY()
            new_dataset = self.dataset[['longitude','latitude','time']]
            new_dataset['absolute_error'] = ae
            new_dataset['error'] = diff
            new_dataset['mae'] = mae
            new_dataset['rmse'] = rmse
            new_dataset['corr'] = corr
            new_dataset['cov'] = cov
            new_object.dataset = new_dataset
            return new_object
        else:
            self.dataset['absolute_error'] = ae
            self.dataset['error'] = diff
            self.dataset['mae'] = mae
            self.dataset['rmse'] = rmse
            self.dataset['corr'] = corr
            self.dataset['cov'] = cov


import logging
import sys
import io
import warnings
import inspect
import traceback


DEFAULT_FORMAT = "%(asctime)s %(levelname)s %(message)s"


def get_logger(
        name: str = None,
        level: int = logging.DEBUG
):
    logger = logging.getLogger(name=name)
    logger.setLevel(level)
    return logger


def create_handler(
        logger: logging.Logger,
        stream: io.TextIOWrapper = sys.stdout,
        format_string: str = DEFAULT_FORMAT
):
    handler = logging.StreamHandler(stream)
    handler.setLevel(logger.level)
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)
    return handler


def setup_logging(
        name: str = None,
        level: int = logging.DEBUG,
        stream: io.TextIOWrapper = sys.stdout,
        format_string: str = DEFAULT_FORMAT
):
    logger = get_logger(name=name, level=level)
    handler = create_handler(logger, stream=stream, format_string=format_string)
    logger.addHandler(handler)

    return logger, handler


def get_slug(
        obj: object
):
    name = obj.__class__.__name__
    ref = hex(id(object))
    return "{0} object at {1}".format(name, ref)


def get_source(level=1):
    frame = inspect.stack()
    if type(frame) is list:
        frame = frame[level]

    file = frame[1]
    line = str(frame[2])
    func = frame[3]

    return file, line, func


def add_info(msg, level=3):
    source = get_source(level=level)
    if isinstance(msg, Exception):
        msg = f"{msg.__class__.__name__}: {str(msg)}\n" + "".join(traceback.format_tb(msg.__traceback__))
    msg = f"{source[0]}.{source[2]}.{source[1]}: {msg}"
    return msg


def debug(msg, *args, **kwargs):
    return logging.debug(add_info(msg), *args, **kwargs)


def info(msg, *args, **kwargs):
    return logging.info(add_info(msg), *args, **kwargs)


def warning(msg, *args, **kwargs):
    return logging.warning(add_info(msg), *args, **kwargs)


def warn(msg, *args, **kwargs):
    return warnings.warn(add_info(msg), *args, **kwargs)


def error(msg, *args, **kwargs):
    return logging.error(add_info(msg), *args, **kwargs)



'''
Python definitions used to aid in the calculation of Continuous Ranked
Probability Score.

*Methods Overview*
    -> crps_sonf_fixed(): Single obs neighbourhood forecast CRPS for fixed obs
    -> crps_song_moving(): Same as above for moving obs
'''

import numpy as np
import xarray as xr
from .CDF import CDF

def crps_sonf_fixed( mod_array, obs_lon, obs_lat, obs_var, obs_time, 
                      nh_radius: float, cdf_type:str, time_interp:str
    ):
    '''
    Handles the calculation of single-observation neighbourhood forecast CRPS
    for a time series at a fixed observation location. Differs from 
    crps_sonf_moving in that it only need calculate a model neighbourhood once.

    Parameters
    ----------
    mod_array   : (xarray DataArray) DataArray from a Model Dataset
    obs_lon     : (float) Longitude of fixed observation point
    obs_lat     : (float) Latitude of fixed observation point
    obs_var     : (array) of floatArray of variable values, e.g time series
    obs_time    : (array) of datetimeArray of times, corresponding to obs_var
    nh_radius   : (float) Neighbourhood radius in km
    cdf_type    : (str) Type of CDF to use for model data. Either 'empirical' 
                   or 'theoretical'.
    time_interp : (str) Type of time interpolation to use

    Returns
    -------
    crps_list     : Array of CRPS values
    n_model_pts   : Array containing the number of model points used for 
                    each CRPS value
    contains_land : Array of bools indicating where a model neighbourhood 
                    contained land.

    '''

    # Define output arrays
    n_neighbourhoods = obs_var.shape[0] 
    crps_list     = np.zeros( n_neighbourhoods )*np.nan
    n_model_pts   = np.zeros( n_neighbourhoods )*np.nan
    contains_land = np.zeros( n_neighbourhoods , dtype=bool)

    # Loop over neighbourhoods
    neighbourhood_indices = np.arange(0,n_neighbourhoods)
    
    # Get model neighbourhood subset using specified method
    subset_ind = subset_indices_by_distance(
                     mod_array.longitude, mod_array.latitude, 
                     obs_lon, obs_lat, nh_radius)
    mod_subset = mod_array.isel(y_dim = subset_ind[0],
                                  x_dim = subset_ind[1])
    mod_subset = mod_subset.swap_dims({'t_dim':'time'})
    
    # Check that the model neighbourhood contains points
    if subset_ind[0].shape[0] == 0 or subset_ind[1].shape[0] == 0:
        crps_list = crps_list*np.nan
    else:
        # Subset model data in time and space: model -> obs
        for ii in neighbourhood_indices:
            
                mod_subset_time = mod_subset.interp(
                                  time = obs_time[ii], method = time_interp,
                                  kwargs={'fill_value':'extrapolate'})
                
                #Check if neighbourhood contains a land value (TODO:mask)
                if any(np.isnan(mod_subset_time)):
                    contains_land[ii] = True
                # Check that neighbourhood contains a value
                if all(np.isnan(mod_subset_time)):
                    pass
                else:
                    # Create model and observation CDF objects
                    mod_cdf = CDF(mod_subset_time, cdf_type = cdf_type)
                
                    # Calculate CRPS and put into output array
                    crps_list[ii] = mod_cdf.crps_fast(obs_var[ii])
                    n_model_pts[ii] = int(mod_subset.shape[0])

    return crps_list, n_model_pts, contains_land

def crps_sonf_moving( mod_array, obs_lon, obs_lat, obs_var, obs_time, 
                      nh_radius: float, cdf_type:str, time_interp:str
    ):
    '''
    Handles the calculation of single-observation neighbourhood forecast CRPS
    for a moving observation instrument. Differs from crps_sonf_fixed in that 
    latitude and longitude are arrays of locations.

    Parameters
    ----------
    mod_array   : (xarray DataArray) DataArray from a Model Dataset
    obs_lon     : (array) Longitudes of fixed observation point
    obs_lat     : (array) Latitudes of fixed observation point
    obs_var     : (array) of floatArray of variable values, e.g time series
    obs_time    : (array) of datetimeArray of times, corresponding to obs_var
    nh_radius   : (float) Neighbourhood radius in km
    cdf_type    : (str) Type of CDF to use for model data. Either 'empirical' 
                   or 'theoretical'.
    time_interp : (str) Type of time interpolation to use

    Returns
    -------
    crps_list     : Array of CRPS values
    n_model_pts   : Array containing the number of model points used for 
                    each CRPS value
    contains_land : Array of bools indicating where a model neighbourhood 
                    contained land.
    '''

    # Define output arrays
    n_neighbourhoods = obs_var.shape[0] 
    crps_list     = np.zeros( n_neighbourhoods )*np.nan
    n_model_pts   = np.zeros( n_neighbourhoods )*np.nan
    contains_land = np.zeros( n_neighbourhoods , dtype=bool)

    # Loop over neighbourhoods
    neighbourhood_indices = np.arange(0,n_neighbourhoods)
    for ii in neighbourhood_indices:
        
        # Neighbourhood centre
        cntr_lon = obs_lon[ii]
        cntr_lat = obs_lat[ii]
    
        # Get model neighbourhood subset using specified method
        subset_ind = subset_indices_by_distance(
                         mod_array.longitude, mod_array.latitude, 
                         cntr_lon, cntr_lat, nh_radius)
        
        # Check that the model neighbourhood contains points
        if subset_ind[0].shape[0] == 0 or subset_ind[1].shape[0] == 0:
            crps_list[ii] = np.nan
        else:
            # Subset model data in time and space: model -> obs
            mod_subset = mod_array.isel(y_dim = subset_ind[0],
                                        x_dim = subset_ind[1])
            mod_subset = mod_subset.swap_dims({'t_dim':'time'})
            mod_subset = mod_subset.interp(
                             time = obs_time[ii], method = time_interp,
                             kwargs={'fill_value':'extrapolate'})
            
            #Check if neighbourhood contains a land value (TODO:mask)
            if any(np.isnan(mod_subset)):
                contains_land[ii] = True
            # Check that neighbourhood contains a value
            if all(np.isnan(mod_subset)):
                pass
            else:
                # Create model and observation CDF objects
                mod_cdf = CDF(mod_subset, cdf_type = cdf_type)
            
                # Calculate CRPS and put into output array
                crps_list[ii] = mod_cdf.crps_fast(obs_var[ii])
                n_model_pts[ii] = int(mod_subset.shape[0])
                
    return crps_list, n_model_pts, contains_land

def subset_indices_by_distance(
        longitude, latitude, centre_lon: float, centre_lat: float, 
        radius: float
    ):
    """
    This method returns a `tuple` of indices within the `radius` of the lon/lat point given by the user.

    Distance is calculated as haversine - see `self.calculate_haversine_distance`

    :param centre_lon: The longitude of the users central point
    :param centre_lat: The latitude of the users central point
    :param radius: The haversine distance (in km) from the central point
    :return: All indices in a `tuple` with the haversine distance of the central point
    """

    # Calculate the distances between every model point and the specified
    # centre. Calls another routine dist_haversine.

    dist = calculate_haversine_distance(centre_lon, centre_lat, 
                                        longitude, latitude)
    indices_bool = dist < radius
    indices = np.where(indices_bool.compute())

    return xr.DataArray(indices[0]), xr.DataArray(indices[1])

def calculate_haversine_distance(lon1, lat1, lon2, lat2, r = 6371.00717):
    '''
    # Estimation of geographical distance using the Haversine function.
    # Input can be single values or 1D arrays of locations. This
    # does NOT create a distance matrix but outputs another 1D array.
    # This works for either location vectors of equal length OR a single loc
    # and an arbitrary length location vector.
    #
    # lon1, lat1 :: Location(s) 1.
    # lon2, lat2 :: Location(s) 2.
    '''

    # Convert to radians for calculations
    lon1 = xr.ufuncs.deg2rad(lon1)
    lat1 = xr.ufuncs.deg2rad(lat1)
    lon2 = xr.ufuncs.deg2rad(lon2)
    lat2 = xr.ufuncs.deg2rad(lat2)

    # Latitude and longitude differences
    dlat = (lat2 - lat1) / 2
    dlon = (lon2 - lon1) / 2

    # Haversine function.
    distance = xr.ufuncs.sin(dlat) ** 2 + xr.ufuncs.cos(lat1) *  \
        xr.ufuncs.cos(lat2) * xr.ufuncs.sin(dlon) ** 2
    distance = 2 * r * xr.ufuncs.arcsin(xr.ufuncs.sqrt(distance))

    return distance



from dask import array
import xarray as xr
import numpy as np
from dask.distributed import Client
import copy
from .logging_util import get_slug, debug, info, warn, warning


def setup_dask_client(
        workers: int = 2,
        threads: int = 2,
        memory_limit_per_worker: str = '2GB'
):
    Client(n_workers=workers, threads_per_worker=threads, memory_limit=memory_limit_per_worker)


class COAsT:
    def __init__(
            self,
            file: str = None,
            chunks: dict = None,
            multiple=False,
            workers: int = 2,  # TODO Do something with this unused parameter
            threads: int = 2,  # TODO Do something with this unused parameter
            memory_limit_per_worker: str = '2GB'  # TODO Do something with this unused parameter
    ):
        debug(f"Creating a new {get_slug(self)}")
        self.dataset = None
        self.earth_raids = 6371.007176  # Radius of the earth in km TODO Could this be module-level?
        self.set_dimension_mapping()
        self.set_variable_mapping()

        if file is None:
            warn(
                "Object created but no file or directory specified: \n"
                "{0} \n"
                "Use COAsT.load() to load a NetCDF file from file path or directory into this object."
                .format(str(self)),
                UserWarning
            )
        else:
            self.load(file, chunks, multiple)
        debug(f"{get_slug(self)} initialised")

    def load(self, file_or_dir, chunks: dict = None, multiple=False):
        """
        Loads a file into a COAsT object's dataset variable using xarray
        
        Args:
            file_or_dir (str)     : file name or directory to multiple files.
            chunks (dict)  : Chunks to use in Dask [default None]
            multiple (bool): If true, load in multiple files from directory.
                             If false load a single file [default False]
        """
        if multiple:
            self.load_multiple(file_or_dir, chunks)
        else:
            self.load_single(file_or_dir, chunks)    

    def __getitem__(self, name: str):
        return self.dataset[name]

    def load_single(self, file, chunks: dict = None):
        """ Loads a single file into COAsT object's dataset variable. """
        info(f"Loading a single file ({file} for {get_slug(self)}")
        self.dataset = xr.open_dataset(file, chunks=chunks)

    def load_multiple(self, directory_to_files, chunks: dict = None):
        """ Loads multiple files from directory into dataset variable. """
        info(f"Loading a directory ({directory_to_files}) for {get_slug(self)}")
        self.dataset = xr.open_mfdataset(
            directory_to_files, chunks=chunks, parallel=True, 
            combine="by_coords") #, compat='override')

    def load_dataset(self, dataset):
        """        
        :param dataset: The dataset to use
        :type dataset: xarray.Dataset
        """
        self.dataset = dataset
        debug(f"Dataset for {get_slug(self)} set to {get_slug(dataset)}")
    
    def set_dimension_mapping(self):
        self.dim_mapping = None
        debug(f"dim_mapping for {get_slug(self)} set to {self.dim_mapping}")
        
    def set_variable_mapping(self):
        self.var_mapping = None
        debug(f"var_mapping for {get_slug(self)} set to {self.var_mapping}")

    def set_grid_ref_attr(self):
        self.grid_ref_attr_mapping = None
        debug(f"grid_ref_attr_mapping for {get_slug(self)} set to {self.grid_ref_attr_mapping}")

    def set_dimension_names(self, dim_mapping: dict):
        """ 
        Relabel dimensions in COAsT object xarray.dataset to ensure
        consistent naming throughout the COAsT package.
        
        Args:
            dim_mapping (dict): keys are dimension names to change and values
                                new dimension names
        """
        debug(f"Setting dimension names for {get_slug(self)} with mapping {dim_mapping}")
        if dim_mapping is None:
            return
        for key, value in dim_mapping.items():
            try:
                self.dataset = self.dataset.rename_dims({ key : value })
            except:  # TODO Catch specific exception(s)
                warning(f"{get_slug(self)}: Problem renaming dimension from {get_slug(self.dataset)}: {key} -> {value}")
                
    def set_variable_names(self, var_mapping: dict):
        """ 
        Relabel variables in COAsT object xarray.dataset to ensure
        consistent naming throughout the COAsT package.
        
        Args:
            var_mapping (dict): keys are variable names to change and values
                                are new variable names
        """
        debug(f"Setting variable names for {get_slug(self)} with mapping {var_mapping}")
        if var_mapping is None:
            return
        for key, value in var_mapping.items():
            try:
                self.dataset = self.dataset.rename_vars({ key : value })
            except:
                warning(f"{get_slug(self)}: Problem renaming variables from {get_slug(self.dataset)}: {key} -> {value}")

    def set_variable_grid_ref_attribute(self, grid_ref_attr_mapping: dict):
        """
        Set attributes for variables to access within package.
        Set grid attributes to identify with grid variable is associated with.
        """
        debug(f"Setting variable attributes for {get_slug(self)} with mapping {grid_ref_attr_mapping}")
        if grid_ref_attr_mapping is None: return
        for key, value in grid_ref_attr_mapping.items():
            try:
                self.dataset[key].attrs['grid_ref'] = value
            except:
                warning(f"{get_slug(self)}: Problem assigning attributes in {get_slug(self.dataset)}: {key} -> {value}")

    def copy(self):
        new = copy.copy(self)
        debug(f"Copied {get_slug(self)} to new {get_slug(new)}")
        return new
        
    def isel(self, indexers: dict = None, drop: bool = False,
             **kwargs):
        '''
        Indexes COAsT object along specified dimensions using xarray isel. 
        Input is of same form as xarray.isel. Basic use, hand in either:
            1. Dictionary with keys = dimensions, values = indices
            2. **kwargs of form dimension = indices
        '''
        obj_copy = self.copy()
        debug(f"Indexing (isel) {get_slug(obj_copy)}")
        obj_copy.dataset = obj_copy.dataset.isel(indexers, drop, **kwargs)
        return obj_copy
    
    def sel(self, indexers: dict = None, drop: bool = False,
             **kwargs):
        '''
        Indexes COAsT object along specified dimensions using xarray sel. 
        Input is of same form as xarray.sel. Basic use, hand in either:
            1. Dictionary with keys = dimensions, values = indices
            2. **kwargs of form dimension = indices
        '''
        obj_copy = self.copy()
        debug(f"Indexing (sel) {get_slug(obj_copy)}")
        obj_copy.dataset = obj_copy.dataset.sel(indexers, drop, **kwargs)
        return obj_copy
    
    def rename(self, rename_dict, inplace: bool=None, **kwargs):
        debug(f"Renaming {get_slug(self.dataset)} with dict {rename_dict}")
        self.dataset = self.dataset.rename(rename_dict, inplace, **kwargs)
        return  # TODO Should this return something? If not, the statement is not needed

    def subset(self, **kwargs):
        '''
        Subsets all variables within the dataset inside self (a COAsT object).
        Input is a set of keyword argument pairs of the form:
            dimension_name = indices
        The entire object is then subsetted along this dimension at indices
        '''
        debug(f"Subsetting {get_slug(self)}")
        self.dataset = self.dataset.isel(kwargs)
        
    def subset_as_copy(self, **kwargs):
        '''
        Similar to COAsT.subset() however applies the subsetting to a copy of
        the original COAsT object. This subsetted copy is then returned.
        Useful for preserving the original object whilst creating smaller
        subsetted object copies.
        '''
        debug(f"Subsetting as copy {get_slug(self.dataset)}")
        obj_copy = self.copy()
        obj_copy.subset(**kwargs)
        return obj_copy

    def distance_between_two_points(self):
        raise NotImplementedError
        
    def subset_indices_by_distance(self, centre_lon: float, centre_lat: float, 
                                   radius: float):
        """
        This method returns a `tuple` of indices within the `radius` of the lon/lat point given by the user.

        Distance is calculated as haversine - see `self.calculate_haversine_distance`

        :param centre_lon: The longitude of the users central point
        :param centre_lat: The latitude of the users central point
        :param radius: The haversine distance (in km) from the central point
        :return: All indices in a `tuple` with the haversine distance of the central point
        """
        debug(f"Subsetting {self} indices by distance")
        # Flatten NEMO domain stuff.
        lon = self.dataset.longitude
        lat = self.dataset.latitude

        # Calculate the distances between every model point and the specified
        # centre. Calls another routine dist_haversine.

        dist = self.calculate_haversine_distance(centre_lon, centre_lat, 
                                                 lon, lat)
        indices_bool = dist < radius
        indices = np.where(indices_bool.compute())

        return xr.DataArray(indices[0]), xr.DataArray(indices[1])
    
    def subset_indices_lonlat_box(self, lonbounds, latbounds):
        """Generates array indices for data which lies in a given lon/lat box.

        Keyword arguments:
        lon       -- Longitudes, 1D or 2D.
        lat       -- Latitudes, 1D or 2D
        lonbounds -- Array of form [min_longitude=-180, max_longitude=180]
        latbounds -- Array of form [min_latitude, max_latitude]
        
        return: Indices corresponding to datapoints inside specified box
        """
        debug(f"Subsetting {get_slug(self)} indices within lon/lat")
        lon_str = 'longitude'
        lat_str = 'latitude'
        lon = self.dataset[lon_str].copy()  # TODO Add a comment explaining why this needs to be copied
        lat = self.dataset[lat_str]
        ff = lon > lonbounds[0]
        ff *= lon < lonbounds[1]
        ff *= lat > latbounds[0]
        ff *= lat < latbounds[1]

        return np.where(ff)

    def calculate_haversine_distance(self, lon1, lat1, lon2, lat2):  # TODO This could be a static method
        '''
        # Estimation of geographical distance using the Haversine function.
        # Input can be single values or 1D arrays of locations. This
        # does NOT create a distance matrix but outputs another 1D array.
        # This works for either location vectors of equal length OR a single loc
        # and an arbitrary length location vector.
        #
        # lon1, lat1 :: Location(s) 1.
        # lon2, lat2 :: Location(s) 2.
        '''

        debug(f"Calculating haversine distance between {lon1},{lat1} and {lon2},{lat2}")

        # Convert to radians for calculations
        lon1 = xr.ufuncs.deg2rad(lon1)
        lat1 = xr.ufuncs.deg2rad(lat1)
        lon2 = xr.ufuncs.deg2rad(lon2)
        lat2 = xr.ufuncs.deg2rad(lat2)

        # Latitude and longitude differences
        dlat = (lat2 - lat1) / 2
        dlon = (lon2 - lon1) / 2

        # Haversine function.
        distance = xr.ufuncs.sin(dlat) ** 2 + xr.ufuncs.cos(lat1) * xr.ufuncs.cos(lat2) * xr.ufuncs.sin(dlon) ** 2
        distance = 2 * 6371.007176 * xr.ufuncs.arcsin(xr.ufuncs.sqrt(distance))

        return distance

    def get_subset_as_xarray(self, var: str, points_x: slice, points_y: slice, line_length: int = None,
                             time_counter: int = 0):
        """
        This method gets a subset of the data across the x/y indices given for the chosen variable.

        Setting time_counter to None will treat `var` as only having 3 dimensions depth, y, x

        there is a check on `var` to see the size of the time_counter, if 1 then time_counter is fixed to index 0.

        :param var: the name of the variable to get data from
        :param points_x: a list/array of indices for the x dimension
        :param points_y: a list/array of indices for the y dimension
        :param line_length: (Optional) the length of your subset (assuming simple line transect)
        :param time_counter: (Optional) which time slice to get data from, if None and the variable only has one a time
                             channel of length 1 then time_counter is fixed too an index of 0
        :return: data across all depths for the chosen variable along the given indices
        """
        debug(f"Subsetting {var} from {get_slug(self)}")
        try:
            [time_size, _, _, _] = self.dataset[var].shape
            if time_size == 1:
                time_counter == 0  # TODO This should probably be =, not ==

        except ValueError:
            time_counter = None

        dx = xr.DataArray(points_x)
        dy = xr.DataArray(points_y)

        if time_counter is None:
            smaller = self.dataset[var].isel(x_dim=dx, y_dim=dy)
        else:
            smaller = self.dataset[var].isel(t_dim=time_counter, x_dim=dx, y_dim=dy)

        return smaller

    def get_2d_subset_as_xarray(self, var: str, points_x: slice, points_y: slice, line_length: int = None,
                                time_counter: int = 0):
        """

        :param var:
        :param points_x:
        :param points_y:
        :param line_length:
        :param time_counter:
        :return:
        """

        debug(f"Fetching {var} subset as xarray")
        try:
            [time_size, _, _, _] = self.dataset[var].shape
            if time_size == 1:
                time_counter == 0
        except ValueError:
            time_counter = None

        if time_counter is None:
            smaller = self.dataset[var].isel(x=points_x, y=points_y)
        else:
            smaller = self.dataset[var].isel(time_counter=time_counter, x=points_x, y=points_y)

        return smaller

    def plot_simple_2d(self, x, y, data: xr.DataArray, cmap, plot_info: dict):
        """
        This is a simple method that will plot data in a 2d. It is a wrapper for matplotlib's 'pcolormesh' method.

        `cmap` and `plot_info` are required to run this method, `cmap` is passed directly to `pcolormesh`.

        `plot_info` contains all the required information for setting the figure;
         - ylim
         - xlim
         - clim
         - title
         - fig_size
         - ylabel

        :param x: The variable contain the x axis information
        :param y: The variable contain the y axis information
        :param data: the DataArray a user wishes to plot
        :param cmap:
        :param plot_info:
        :return:
        """
        info("Generating simple 2D plot...")
        import matplotlib.pyplot as plt

        plt.close('all')

        fig = plt.figure()
        plt.rcParams['figure.figsize'] = plot_info['fig_size']

        ax = fig.add_subplot(411)
        plt.pcolormesh(x, y, data, cmap=cmap)

        plt.ylim(plot_info['ylim'])
        plt.xlim(plot_info['xlim'])
        plt.title(plot_info['title'])
        plt.ylabel(plot_info['ylabel'])
        plt.clim(plot_info['clim'])
        plt.colorbar()

        return plt

    def plot_cartopy(self, var: str, plot_var: array, params, time_counter: int = 0):
        try:
            import cartopy.crs as ccrs  # mapping plots
            import cartopy.feature  # add rivers, regional boundaries etc
            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
            from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
        except ImportError:
            import sys
            warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
            sys.exit(-1)

        import matplotlib.pyplot as plt

        info("Generating CartoPy plot...")
        plt.close('all')
        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca()
        ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())

        cset = self.dataset[var].isel(time_counter=time_counter, deptht=0).plot.pcolormesh \
            (np.ma.masked_where(plot_var == np.NaN, plot_var), transform=ccrs.PlateCarree(), cmap=params.cmap)

        cset.set_clim([params.levs[0], params.levs[-1]])

        ax.add_feature(cartopy.feature.OCEAN)
        ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
        ax.add_feature(cartopy.feature.RIVERS)
        coast = NaturalEarthFeature(category='physical', scale='10m', facecolor='none', name='coastline')
        ax.add_feature(coast, edgecolor='gray')

        gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                          linewidth=0.5, color='gray', alpha=0.5, linestyle='-')

        gl.xlabels_top = False
        gl.xlabels_bottom = True
        gl.ylabels_right = False
        gl.ylabels_left = True
        gl.xformatter = LONGITUDE_FORMATTER
        gl.yformatter = LATITUDE_FORMATTER

        plt.colorbar(cset, shrink=params.colorbar_shrink, pad=.05)

        # tmp = self.dataset.votemper
        # tmp.attrs = self.dataset.votemper.attrs
        # tmp.isel(time_counter=time_counter, deptht=0).plot.contourf(ax=ax, transform=ccrs.PlateCarree())
        # ax.set_global()
        # ax.coastlines()
        info("Displaying plot!")
        plt.show()

    def plot_movie(self):
        raise NotImplementedError



from .COAsT import COAsT
from .COAsT import setup_dask_client
from .NEMO import NEMO
from .TRANSECT import Transect
from .ALTIMETRY import ALTIMETRY
from .OBSERVATION import OBSERVATION
from .CDF import CDF
from .INTERNALTIDE import INTERNALTIDE
from .TIDEGAUGE import TIDEGAUGE
from . import logging_util
from . import general_utils
from . import plot_util
from . import crps_util
from .CONTOUR import Contour, Contour_f, Contour_t
